from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType   
from typing import List, Callable
from rb.similarity.vector_model import VectorModel
from rb.cna.cna_graph import CnaGraph
from rb.cna.edge_type import EdgeType

from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class MiddleEndCohesion(ComplexityIndex):

    """ only between mid block and end block """
    def __init__(self, lang: Lang,
            reduce_depth: int = None, reduce_function: MeasureFunction = None,
            cna_graph: CnaGraph = None):
        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.COHESION,
                                 reduce_depth=reduce_depth, reduce_function=reduce_function,
                                 abbr="MidEndCohesion")
        self.cna_graph = cna_graph
        
    def process(self, element: TextElement) -> float:
        return self.compute(element)

    def compute(self, element: TextElement) -> float:

        blocks = element.get_blocks()
        if len(blocks) < 3:
            element.indices[self] = ComplexityIndex.IDENTITY
            return ComplexityIndex.IDENTITY            
        else:
            end_block = blocks[-1]
            end_index = len(blocks) - 1
            scale_factor, weighted_sum = 0, 0

            for i, _ in enumerate(blocks):
                if i == end_index:  continue
                sim_edge = self.cna_graph.edges(node=(end_block, blocks[i]), edge_type=EdgeType.SEMANTIC,
                                                vector_model=None)
                v = sim_edge[0][2]
                weighted_sum += v * (1.0 / (end_index - i))
                scale_factor += (1.0 / (end_index -  i))
            element.indices[self] = weighted_sum / scale_factor 
            return element.indices[self]


import json
from typing import List, Dict
import csv
from dateutil import parser
from datetime import datetime
import xmltodict

from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.core.cscl.contribution import Contribution
from rb.core.cscl.conversation import Conversation
from rb.core.cscl.community import Community
from rb.core.cscl.cscl_indices import CsclIndices
from rb.cna.cna_graph import CnaGraph
from rb.similarity.vector_model_factory import create_vector_model
from rb.similarity.vector_model import VectorModelType

from rb.utils.rblogger import Logger
from rb.processings.cscl.community_processing import compute_sna_metrics, determine_participant_contributions, determine_participation
from rb.processings.cscl.participant_evaluation import evaluate_interaction, evaluate_involvement, evaluate_used_concepts, perform_sna

logger = Logger.get_logger()

CONTRIBUTIONS_KEY = 'contributions'
ID_KEY = 'id'
PARENT_ID_KEY = 'parent_id'
TIMESTAMP_KEY = 'timestamp'
TEXT_KEY = 'text'
USER_KEY = 'user'

def get_json_from_json_file(filename: str) -> Dict:
	conversation_thread = dict()
	contribution_list = []

	with open(filename, "rt", encoding='utf-8') as json_file:
		contribution_list_json = json.load(json_file)
		for cnt in contribution_list_json:
			contribution = {
				ID_KEY: cnt['genid'],
				PARENT_ID_KEY: cnt["ref"],
				TIMESTAMP_KEY: int(float(cnt['time'])),
				USER_KEY: cnt["nickname"],
				TEXT_KEY: cnt["text"],
			}
			contribution_list.append(contribution)

	conversation_thread[CONTRIBUTIONS_KEY] = contribution_list

	return conversation_thread

def get_json_from_csv(filename: str) -> Dict:
	conversation_thread = dict()
	contribution_list = []

	with open(filename) as csv_file:
		csv_reader = csv.reader(csv_file, delimiter=',')
		first = True			

		for row in csv_reader:
			if first:
				first = False
			else:
				contribution = dict()

				contribution[ID_KEY] = row[0]
				contribution[PARENT_ID_KEY] = row[1]
				contribution[USER_KEY] = row[2]
				contribution[TEXT_KEY] = row[3]
				contribution[TIMESTAMP_KEY] = row[4]

				contribution_list.append(contribution)

	conversation_thread[CONTRIBUTIONS_KEY] = contribution_list

	return conversation_thread

def parse_large_csv(filename: str) -> Dict:
	conversation_thread = dict()
	contribution_list = []

	with open(filename) as csv_file:
		csv_reader = csv.reader(csv_file, delimiter=',')
		first = True			

		for row in csv_reader:
			if first:
				first = False
			else:
				contribution = dict()

				contribution[ID_KEY] = row[0]
				contribution[PARENT_ID_KEY] = row[1]
				if row[1] == '':
					contribution[PARENT_ID_KEY] = '-1'

				contribution[USER_KEY] = row[2]
				contribution[TEXT_KEY] = row[6]

				contribution[TIMESTAMP_KEY] = datetime.timestamp(parser.parse(row[3], ignoretz=True, fuzzy=True))

				contribution_list.append(contribution)

	conversation_thread[CONTRIBUTIONS_KEY] = contribution_list

	return conversation_thread

def load_from_xml(lang: Lang, filename: str) -> Dict:
	with open(filename, "rt") as f:
		my_dict=xmltodict.parse(f.read())
		contributions = [
			{
				ID_KEY: int(utterance["@genid"]) - 1,
				PARENT_ID_KEY: int(utterance["@ref"]) - 1,
				TIMESTAMP_KEY: datetime.timestamp(datetime.strptime(utterance["@time"],'%H.%M.%S').replace(year=2020)),
				USER_KEY: turn["@nickname"],
				TEXT_KEY: utterance["#text"],
			}
			for turn in my_dict["corpus"]["Dialog"]["Body"]["Turn"]
			for utterance in (turn["Utterance"] if isinstance(turn["Utterance"], List) else [turn["Utterance"]])
		]

		return {CONTRIBUTIONS_KEY: contributions}


def compute_indices(conv: Conversation):
	participant_list = conv.get_participants()
	names = list(map(lambda p: p.get_id(), participant_list))

	print('Participants are:')
	print(names)

	print('Begin computing indices')

	evaluate_interaction(conv)
	evaluate_involvement(conv)
	evaluate_used_concepts(conv)
	perform_sna(conv, False)

	print('Finished computing indices')

	for p in participant_list:
		print('Printing for participant ' + p.get_id())

		print("SCORE  " + str(p.get_index(CsclIndices.SCORE)))
		print("NO_NOUNS " + str(p.get_index(CsclIndices.NO_NOUNS)))
		print("NO_VERBS " + str(p.get_index(CsclIndices.NO_VERBS)))
		print("NO_CONTRIBUTION " + str(p.get_index(CsclIndices.NO_CONTRIBUTION)))
		print("SOCIAL_KB " + str(p.get_index(CsclIndices.SOCIAL_KB)))
		print("INDEGREE " + str(p.get_index(CsclIndices.INDEGREE)))
		print("OUTDEGREE " + str(p.get_index(CsclIndices.OUTDEGREE)))

		print('---------------------')

	for n1 in names:
		for n2 in names:
			print('Score for ' + n1 + ' ' + n2 + ' is:')
			print(conv.get_score(n1, n2))


def test_community_processing(community: Community):
	print('Testing Community Processing')

	participant_list = community.get_participants()
	names = list(map(lambda p: p.get_id(), participant_list))

	print('Participants are:')
	print(names)

	print('Begin computing indices')

	determine_participant_contributions(community)
	determine_participation(community)
	compute_sna_metrics(community)

	print('Finished computing indices')

	for p in participant_list:
		print('Printing for participant ' + p.get_id())

		print(p.get_index(CsclIndices.SCORE))
		print(p.get_index(CsclIndices.NO_NOUNS))
		print(p.get_index(CsclIndices.NO_VERBS))
		print(p.get_index(CsclIndices.NO_CONTRIBUTION))
		print(p.get_index(CsclIndices.SOCIAL_KB))
		print(p.get_index(CsclIndices.INDEGREE))
		print(p.get_index(CsclIndices.OUTDEGREE))
		print(p.get_index(CsclIndices.NO_NEW_THREADS))
		print(p.get_index(CsclIndices.NEW_THREADS_OVERALL_SCORE))
		print(p.get_index(CsclIndices.NEW_THREADS_CUMULATIVE_SOCIAL_KB))
		print(p.get_index(CsclIndices.AVERAGE_LENGTH_NEW_THREADS))


		print('---------------------')

	for n1 in names:
		for n2 in names:
			print('Score for ' + n1 + ' ' + n2 + ' is:')
			print(community.get_score(n1, n2))

def main():
	# Test community processing - English discussion csv
	conv_thread = []
	for i in range(1, 20):
		conversation = get_json_from_json_file("./jsons/conversation_" + str(i) + ".json")
		conv_thread.append(conversation)

	community = Community(lang=Lang.EN, container=None, community=conv_thread)
	en_coca_word2vec = create_vector_model(Lang.EN, VectorModelType.from_str("word2vec"), "coca")
	community.graph = CnaGraph(docs=[community], models=[en_coca_word2vec])

	conv = community.get_conversations()[0]
	conv.container.graph = CnaGraph(docs=[conv], models=[en_coca_word2vec])
	compute_indices(conv)
	# test_community_processing(community)

	# Test for English discussion CSV

	'''
	print('Testing English CSV')

	conv_thread = parse_large_csv('./thread.csv')

	community = Community(lang=Lang.EN, container=None, community=[conv_thread])
	en_coca_word2vec = create_vector_model(Lang.EN, VectorModelType.from_str("word2vec"), "coca")
	community.graph = CnaGraph(docs=[community], models=[en_coca_word2vec])

	conv = community.get_conversations()[0]

	compute_indices(conv)
	'''

	# Test for French discussion XML

	'''

	print('Testing French XML')

	conv_thread = load_from_xml(Lang.FR, './conpa-MEEF-anonyme.xml')
	
	community = Community(lang=Lang.FR, container=None, community=[conv_thread])
	fr_coca_word2vec = create_vector_model(Lang.FR, VectorModelType.from_str("word2vec"), "coca")
	community.graph = CnaGraph(docs=[community], models=[fr_coca_word2vec])

	conv = community.get_conversations()[0]

	compute_indices(conv)

	'''

if __name__ == '__main__':
	main()



from typing import List, Dict

from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.block import Block
from rb.core.pos import POS
from rb.core.cscl.contribution import Contribution
from rb.core.cscl.conversation import Conversation
from rb.cna.cna_graph import CnaGraph
from rb.core.cscl.cscl_indices import CsclIndices

from rb.utils.rblogger import Logger

def get_block_importance(block_importance: Dict[Block, Dict[Block, float]], a: Block, b: Block) -> float:
	if not (a in block_importance):
		return 0
	if not (b in block_importance[a]):
		return 0

	return block_importance[a][b]


def evaluate_interaction(conversation: Conversation):
	cna_graph = conversation.container.graph
	importance = cna_graph.importance
	block_importance = cna_graph.block_importance

	participants = conversation.get_participants()
	contributions = conversation.get_contributions()
	conversation.init_scores()

	if (len(participants) == 0):
		return

	for i, contribution1 in enumerate(contributions):
		p1 = contribution1.get_participant().get_id()

		conversation.update_score(p1, p1, importance[contribution1])

		for j in range(0, i):
			contribution2 = contributions[j]

			if get_block_importance(block_importance, contribution1, contribution2) > 0:
				p2 = contribution2.get_participant().get_id()

				added_kb = importance[contribution1] * get_block_importance(block_importance,
																				contribution1, contribution2)

				conversation.update_score(p1, p2, added_kb)


def evaluate_involvement(conversation: Conversation):
	cna_graph = conversation.container.graph
	importance = cna_graph.importance
	block_importance = cna_graph.block_importance
	participants = conversation.get_participants()

	if (len(participants) == 0):
		return

	for contribution in conversation.get_contributions():
		p = contribution.get_participant()

		current_value = p.get_index(CsclIndices.SCORE)
		p.set_index(CsclIndices.SCORE, current_value + importance[contribution])

		current_value = p.get_index(CsclIndices.SOCIAL_KB)
		parent_contribution = contribution.get_parent()

		if parent_contribution != None:
			current_value += (get_block_importance(block_importance, contribution, parent_contribution) *
								importance[contribution])

			p.set_index(CsclIndices.SOCIAL_KB, current_value)

		current_value = p.get_index(CsclIndices.NO_CONTRIBUTION)
		p.set_index(CsclIndices.NO_CONTRIBUTION, current_value + 1)

	
def evaluate_used_concepts(conversation: Conversation):
	participants = conversation.get_participants()
	
	for p in participants:
		for contribution in conversation.get_participant_contributions(p.get_id()):
			for word in contribution.get_words():
				if word.pos.to_wordnet() == 'n':
					current_value = p.get_index(CsclIndices.NO_NOUNS)
					p.set_index(CsclIndices.NO_NOUNS, current_value + 1)
				if word.pos.to_wordnet() == 'v':
					current_value = p.get_index(CsclIndices.NO_VERBS)
					p.set_index(CsclIndices.NO_VERBS, current_value + 1)


def perform_sna(conversation: Conversation, needs_anonymization: bool):
	participants = conversation.get_participants()

	for i, p1 in enumerate(participants):
		for j, p2 in enumerate(participants):
			if i != j:
				current_value = p1.get_index(CsclIndices.OUTDEGREE)
				p1.set_index(CsclIndices.OUTDEGREE, current_value +
											conversation.get_score(p1.get_id(), p2.get_id()))

				current_value = p2.get_index(CsclIndices.INDEGREE)
				p2.set_index(CsclIndices.INDEGREE, current_value +
											conversation.get_score(p1.get_id(), p2.get_id()))
			else:
				current_value = p1.get_index(CsclIndices.OUTDEGREE)
				p1.set_index(CsclIndices.OUTDEGREE, current_value +
											conversation.get_score(p1.get_id(), p1.get_id()))






from rb.core.lang import Lang
from rb.core.text_element_type import TextElementType
import numpy as np
from typing import List, Dict

from rb.core.lang import Lang


class TextElement:
    """General class of TextElement
    Each node from the document's tree is a TextElement

    Attributes
    ----------
    lang : Lang
        language where the index is applied
    text : str
        raw text of the element
    depth : int
        Depth in the document's tree. It is used to determine what type of element it is. 
        depth 0 means is a word, depth 1 means sentence
        It can increases as much as you want, the last elements
        in the tree being always of the type Document
    container: parent textelement
    """

    def __init__(self, lang: Lang, text: str,
                 depth: int,
                 container: 'TextElement' = None):
        from rb.complexity.complexity_index import ComplexityIndex

        self.text = None
        if text != None:
            self.text = text.strip()

        self.lang = lang
        self.container = container
        self.vectors = {}
        self.components: List[TextElement] = []
        self.vectors_initialized = False
        self.indices: Dict[ComplexityIndex, float] = {}
        self.depth = depth
    
    def is_community(self) -> bool:
        return self.depth >= TextElementType.COMM.value

    def is_conversation(self) -> bool:
        return self.depth >= TextElementType.CONV.value

    def is_contribution(self) -> bool:
        return self.depth == TextElementType.DOC.value

    def is_document(self) -> bool:
        return self.depth >= TextElementType.DOC.value

    def is_block(self) -> bool:
        return self.depth ==  TextElementType.BLOCK.value
    
    def is_sentence(self) -> bool:
        return self.depth == TextElementType.SENT.value
    
    def is_word(self) -> bool:
        return self.depth == TextElementType.WORD.value
        
    def get_parent_document(self) -> "Document":
        if self.is_document():
            return self
        else:
            return self.container.get_parent_document()

    def get_vector(self, model: 'VectorModel') -> np.array:
        return self.vectors[model]

    def get_words(self) -> List["Word"]:
        if self.depth == TextElementType.WORD.value:
            return [self]
        return [word for child in self.components for word in child.get_words()]

    def __eq__(self, other):
        if isinstance(other, TextElement):
            return self.depth == self.depth and self.text == other.text
        return NotImplemented

    def __hash__(self):
        return hash((self.depth, self.text))

    def __iter__(self):
        return iter(self.components)

    def __str__(self):
        return self.text

    def __getstate__(self):
        return {
            key: value
            for key, value in self.__dict__.items()
            if key != "cna_graph"
        }
        
    def __setstate__(self, state):
        for key, value in state.items():
            setattr(self, key, value)

    def get_sentences(self) -> List["Sentence"]:
        if self.depth == TextElementType.SENT.value:
            return [self]

        return  [sent for child in self.components for sent in child.get_sentences()]



from enum import Enum, auto
from rb.core.pos_features.pos_feature import POSFeature
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.lang import Lang
from rb.core.pos import POS
from typing import List
import re


class RoPronTypeEnum(Enum):
    ART = auto()
    DEM = auto()
    EMP = auto()
    IND = auto()
    INT = auto()
    REL = auto()
    NEG = auto()
    PRS = auto()
    TOT = auto()


class RoPOSFeaturePronType(POSFeature):


    _INSTANCE = None

    def __init__(self):
        POSFeature.__init__(self, lang=Lang.RO, name=RoFeaturesName.PRONT,
                            values=None, pos_supported=None)
        self.values = set([ro_pron_type for _, ro_pron_type in RoPronTypeEnum.__members__.items()])
        self.pos_supported = set([POS.DET, POS.PRON, POS.ADV, POS.NUM])

    @classmethod
    def get_instance(cls) -> "RoPOSFeaturePronType":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeaturePronType()
        return cls._INSTANCE

    def get_values(self, tag: str) -> List[RoPronTypeEnum]:
        pattern = '(PronType=[a-zA-Z0-9,]*){1}'
        matches = re.findall(pattern, tag)
        if len(matches) == 0: return []
        nrs = matches[0].split('=')[1].split(',')
        values = []
        for nr in nrs:
            if nr == 'Art':
                values.append(RoPronTypeEnum.ART)
            elif nr == 'Dem':
                values.append(RoPronTypeEnum.DEM)
            elif nr == 'Emp':
                values.append(RoPronTypeEnum.EMP)
            elif nr == 'Int':
                values.append(RoPronTypeEnum.INT)
            elif nr == 'Ind':
                values.append(RoPronTypeEnum.IND)
            elif nr == 'Rel':
                values.append(RoPronTypeEnum.REL)
            elif nr == 'Neg':
                values.append(RoPronTypeEnum.NEG)
            elif nr == 'Prs':
                values.append(RoPronTypeEnum.PRS)
            elif nr == 'Tot':
                values.append(RoPronTypeEnum.TOT)

        return values


import os

from gensim.models import KeyedVectors

from rb.core.lang import Lang
from rb.similarity.vector_model import VectorModel, VectorModelType
from rb.similarity.vector import Vector
from rb.utils.downloader import download_model, check_version
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class Word2Vec(VectorModel):

    def __init__(self, name: str, lang: Lang, dim: int = 300):
        VectorModel.__init__(self, VectorModelType.WORD2VEC, name, lang, dim)  
        

    def load_vectors(self):
        self.load_vectors_from_txt_file("resources/{}/models/{}/word2vec-{}.txt".format(self.lang.value, self.corpus, self.size))
        
        


import os
from typing import List, Tuple, Iterable, Union

import bert
import numpy as np
import tensorflow as tf
import tensorflow_hub as hub
from bert.tokenization.bert_tokenization import FullTokenizer
from rb.core.lang import Lang
from rb.utils.downloader import check_version, download_model
from tensorflow import keras
from transformers import FlaubertTokenizer, TFFlaubertModel
import json


class BertWrapper:

    def __init__(self, lang: Lang, model_name: str = None, max_seq_len: int = 256):
        self.lang = lang
        if lang is Lang.EN:
            if model_name is None:
                model_name = "https://tfhub.dev/tensorflow/bert_en_uncased_L-12_H-768_A-12/1"
            self.bert_layer = hub.KerasLayer(model_name, trainable=True)
            vocab_file = self.bert_layer.resolved_object.vocab_file.asset_path.numpy()
            do_lower_case = self.bert_layer.resolved_object.do_lower_case.numpy()
            self.tokenizer = FullTokenizer(vocab_file, do_lower_case)
        elif lang is Lang.RO:
            if model_name is None:
                model_name = "base"
            if check_version(Lang.RO, ["bert", model_name]):
                download_model(Lang.RO, ["bert", model_name])
            self.model_dir = os.path.join("resources/ro/bert/", model_name)
            json_config_file = os.path.join(self.model_dir, "bert_config.json")
            config = json.load(open(json_config_file, 'r'))
            do_lower_case = bool(config.get('do_lower_case', 0))
            do_remove_accents = bool(config.get('do_remove_accents', 1))
            self.hidden_size = config.get('hidden_size')
            self.tokenizer = FullTokenizer(vocab_file=os.path.join(self.model_dir, "vocab.vocab"), do_lower_case=do_lower_case)
            if do_remove_accents == False:
                self.tokenizer.basic_tokenizer._run_strip_accents = lambda x:x
            bert_params = bert.params_from_pretrained_ckpt(self.model_dir)
            self.bert_layer = bert.BertModelLayer.from_params(bert_params, name="bert_layer")
        elif lang is Lang.FR:
            self.tokenizer = FlaubertTokenizer.from_pretrained("jplu/tf-flaubert-base-cased")
            self.bert_layer = TFFlaubertModel.from_pretrained("jplu/tf-flaubert-base-cased")
            self.bert_layer.call = tf.function(self.bert_layer.call)

        
        self.max_seq_len = max_seq_len
        
        
    def load_weights(self):
        if self.lang is Lang.RO:
            bert.load_bert_weights(self.bert_layer, os.path.join(self.model_dir, "bert_model.ckpt"))

    def create_inputs(self) -> List[keras.layers.Layer]:
        input_ids = tf.keras.layers.Input(shape=(self.max_seq_len,), dtype=tf.int32, name="input_ids")
        segment_ids = tf.keras.layers.Input(shape=(self.max_seq_len,), dtype=tf.int32, name="segment_ids")
        if self.lang is Lang.RO:
            return [input_ids, segment_ids]
        mask_ids = tf.keras.layers.Input(shape=(self.max_seq_len,), dtype=tf.int32, name="mask_ids")
        return [input_ids, mask_ids, segment_ids]
        

    def create_inputs_and_model(self) -> Tuple[List[keras.layers.Layer], tf.Tensor]:
        inputs = self.create_inputs()
        if self.lang is Lang.FR:
            output = self.bert_layer(inputs=inputs[0], attention_mask=inputs[1], token_type_ids=inputs[2])
        else:
            output = self.bert_layer(inputs)
        return inputs, output

    def get_output(self, bert_tensor: tf.Tensor, mode: str = "cls") -> tf.Tensor:
        if self.lang is Lang.EN:
            sequence_output = bert_tensor[1]
        elif self.lang is Lang.RO:
            sequence_output = bert_tensor
        elif self.lang is Lang.FR:
            sequence_output = bert_tensor[0]
        if mode == "cls":
            return tf.keras.layers.Lambda(lambda x: x[:,0,:])(sequence_output)
        elif mode == "pool":
            return tf.keras.layers.GlobalAveragePooling1D()(sequence_output)
        elif mode == "seq":
            return sequence_output
        else:
            print("Unrecognized mode {}".format(mode))
        return None

    def __get_segments(self, tokens):
        sep_enc = self.tokenizer.convert_tokens_to_ids(['[SEP]'])
        segments = []
        current_segment_id = 0
        for token in tokens:
            segments.append(current_segment_id)
            if token == sep_enc[0]:
                if current_segment_id == 0:
                    current_segment_id = 1
        return segments

    def process_input(self, dataset: Iterable[Union[Tuple[str, str], str]]) -> List[np.ndarray]:
        return [np.array(x) for x in zip(*[self.process_text(row) for row in dataset])]

    # use it with sentence2=None for single sentence
    def process_text(self, text: Union[str, Tuple[str, str]]):
        if isinstance(text, str):
            sentence1 = text
            sentence2 = None
        else:
            sentence1, sentence2 = text
        tokens = ['[CLS]']
        tokens.extend(self.tokenizer.tokenize(sentence1))
        tokens.append('[SEP]')
        if sentence2 != None:
            tokens.extend(self.tokenizer.tokenize(sentence2))
            tokens.append('[SEP]')
        input_ids = self.tokenizer.convert_tokens_to_ids(tokens)
        input_masks = [1] * len(input_ids)
        input_segments = self.__get_segments(input_ids)

        if self.max_seq_len == None:
            input_ids = np.array(input_ids)
            input_masks = np.array(input_masks)
            input_segments = np.array(input_segments)
        else: # pad or trim
            if len(input_ids) < self.max_seq_len: # pad
                to_add = self.max_seq_len-len(input_ids)
                for _ in range(to_add):
                    input_ids.append(0)
                    input_masks.append(0)
                    input_segments.append(0)

            elif len(input_ids) > self.max_seq_len: # trim
                if sentence2 is not None:
                    len1 = 0
                    for segment in input_segments:
                        if segment == 1:
                            break
                        len1 += 1
                    len2 = len(input_segments) - len1
                    each = self.max_seq_len // 2
                    if len1 <= each:
                        input_ids = input_ids[:len1] + input_ids[len1:self.max_seq_len - 1] + [input_ids[len1 + len2 - 1]]
                        input_masks = input_masks[:len1] + input_masks[len1:self.max_seq_len - 1] + [input_masks[len1 + len2 - 1]]
                        input_segments = input_segments[:len1] + input_segments[len1:self.max_seq_len - 1] + [input_segments[len1 + len2 - 1]]
                    elif len2 <= each - 1 + self.max_seq_len % 2:
                        input_ids = input_ids[:(self.max_seq_len - len2 - 1)] + input_ids[(len1-1):]
                        input_masks = input_masks[:(self.max_seq_len - len2 - 1)] + input_masks[(len1-1):]
                        input_segments = input_segments[:(self.max_seq_len - len2 - 1)] + input_segments[(len1-1):]
                    else:
                        input_ids = input_ids[:each] + input_ids[(len1 - 1):(len1 + each - 2 + self.max_seq_len % 2)] + [input_ids[-1]]
                        input_masks = input_masks[:each] + input_masks[(len1 - 1):(len1 + each - 2 + self.max_seq_len % 2)] + [input_masks[-1]]
                        input_segments = input_segments[:each] + input_segments[(len1 - 1):(len1 + each - 2 + self.max_seq_len % 2)] + [input_segments[-1]]
                else:
                    input_ids = input_ids[:self.max_seq_len]
                    input_masks = input_masks[:self.max_seq_len]
                    input_segments = input_segments[:self.max_seq_len]

        if self.lang is Lang.RO:
            return input_ids, input_segments
        else:
            return input_ids, input_masks, input_segments

# only used for testing purposes
def test_implementation():

    max_seq_length = 10
    if check_version(Lang.RO, ["bert", "ro0"]):
        download_model(Lang.RO, ["bert", "ro0"])
    # bert_model = BertLayer("resources/ro/bert/ro0/", max_seq_length)
    bert_model = lambda x:x

    input_ids = tf.keras.layers.Input(shape=(max_seq_length,), dtype=tf.int32, name="input_ids")
    segment_ids = tf.keras.layers.Input(shape=(max_seq_length,), dtype=tf.int32, name="segment_ids")
    
    bert_output = bert_model([input_ids, segment_ids])
    print(bert_output.shape)

    model = keras.Model(inputs=[input_ids, segment_ids], outputs=bert_output)
    # model.build(input_shape=[(None, bert_model.max_seq_len), (None, bert_model.max_seq_len)])
    # load pretrained
    # bert.load_bert_weights(bert_model.bert_layer, "resources/ro/bert/ro0/bert_model.ckpt")

    # cls_output = keras.layers.Lambda(lambda seq: seq[:, 0, :])(bert_output)
    # logits = keras.layers.Dense(units=1, activation="softmax")(cls_output)

    s1 = "dar se pare ca nu"
    s2 = "how are you?"

    ids, seg = bert_model.process_sentences(sentence1=s1, sentence2=None)
    # print(ids)
    a = model([np.array([ids]), np.array([seg])])
    print(a)
    bert_model.load_weights()
    a = model([np.array([ids]), np.array([seg])])
    print(a)



#pylint: disable=import-error
import tensorflow as tf
import tensorflow.keras as keras
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Dense, Concatenate, Convolution1D, GlobalMaxPooling1D, Embedding, Dropout, Lambda, Flatten, Add
from tensorflow.keras.layers import Conv1D
from tensorflow.keras.callbacks import TensorBoard
import sys
import numpy as np
from tensorflow.keras import backend as K
import rb.processings.diacritics.utils as utils
import functools

class BertCNN(object):
	"""
	Class to implement simple Bert + Character Level Convolutional Neural Network
	The model is used to classify diacritics
    """


	def __init__(self, window_size, alphabet_size, embedding_size, conv_layers, fc_hidden_size, num_of_classes, batch_max_sentences, batch_max_windows,
				 bert_trainable, cnn_dropout_rate, bert_wrapper, learning_rate, init_model, optimizer='adam', loss='categorical_crossentropy'):
		"""
        Initialization for the Bert + Character Level CNN model.
        Args:
			CNN side
            window_size (int): Size of window
            alphabet_size (int): Size of alphabets to create embeddings for
            embedding_size (int): Size of embeddings
            conv_layers (list[list[int]]): List of Convolution layers for model
            cnn_dropout_rate (float): Dropout Rate for CNN
            
			Bert side
			bert_wrapper (obj): Bert wrapper
			bert_trainable (bool): Whether to train BERT or not
			batch_max_sentences (int): Maximum sentences in batch
			batch_max_windows (int): Maximum windows in batch
			

			init_model (string): Name of model to start training from
			fc_hidden_size (int): Size of hidden layer between features and prediction

			num_of_classes (int): Number of classes in data
			optimizer (str): Training optimizer
            loss (str): Loss function
			learning_rate (float): Learning rate to use for training

        """
		self.window_size = window_size
		self.alphabet_size = alphabet_size
		self.embedding_size = embedding_size
		self.conv_layers = conv_layers
		self.total_number_of_filters = functools.reduce(lambda x,y: x+y[0], conv_layers, 0)
		self.num_of_classes = num_of_classes
		self.cnn_dropout_rate = cnn_dropout_rate
		self.learning_rate = learning_rate
		self.fc_hidden_size = fc_hidden_size

		self.bert_wrapper = bert_wrapper
		self.bert_wrapper.bert_layer.trainable = bert_trainable
		self.batch_max_sentences = batch_max_sentences
		self.batch_max_windows = batch_max_windows

		self.init_model = init_model

		if optimizer == "adam":
			self.optimizer = keras.optimizers.Adam(lr=self.learning_rate)

		if loss == "categorical_crossentropy":
			self.loss = keras.losses.CategoricalCrossentropy(from_logits=False)

		self._build_model()  # builds self.model variable

	
	def _build_embedding_mask(self):
		embedding_mask_weights = np.ones((self.alphabet_size, self.num_of_classes))
		# a -> a, ă, â
		embedding_mask_weights[2] = [1,1,1,0,0]
		# s -> s, ș
		embedding_mask_weights[10] = [1,0,0,1,0]
		# t -> t, ț
		embedding_mask_weights[13] = [1,0,0,0,1]
		# i -> i, î
		embedding_mask_weights[16] = [1,0,1,0,0]
		return embedding_mask_weights
	

	def _build_model(self):
		"""
        Build and compile the Bert + Character Level CNN model
        Returns: None
        """

		# Input layers
		input_bert_ids = Input(shape=(self.batch_max_sentences, self.bert_wrapper.max_seq_len), name='bert_input_ids', dtype='int32')
		input_bert_seg = Input(shape=(self.batch_max_sentences, self.bert_wrapper.max_seq_len), name='bert_segment_ids', dtype='int32')
		input_token_ids = Input(shape=(self.batch_max_windows,), name='token_ids', dtype='int32')
		input_sent_ids = Input(shape=(self.batch_max_windows,), name='sent_ids', dtype='int32')
		input_mask = Input(shape=(self.batch_max_windows,), name='mask', dtype='float32')
		input_char_windows = Input(shape=(self.batch_max_windows, self.window_size), name='char_windows', dtype='int32')
		
		keras_internal_batch_size = K.shape(input_token_ids)[0]

		##########################################################################
		######################  Bert  ############################################
		input_bert_ids_reshaped = tf.reshape(input_bert_ids, shape=(-1, self.bert_wrapper.max_seq_len), name="reshape_input_bert_ids")
		input_bert_seg_reshaped = tf.reshape(input_bert_seg, shape=(-1, self.bert_wrapper.max_seq_len), name="reshape_input_bert_seg")
		# shape = (?batch_size x max_sent, max_seq_len)

		bert_output = self.bert_wrapper.bert_layer(input_bert_ids_reshaped, input_bert_seg_reshaped)
		# bert_output = (?batch_size x max_sent, bert_max_seq_len, bert_hidden_size)
		bert_output = tf.reshape(bert_output, shape=(-1, self.batch_max_sentences, self.bert_wrapper.max_seq_len, self.bert_wrapper.hidden_size), name="bert_output")
		# bert_output = (?batch_size, max_sent, bert_max_seq_len, bert_hidden_size)

		##########################################################################

		##########################################################################
		######################  CharCNN  #########################################
		embedding_mask_weights = self._build_embedding_mask()
		input_char_windows_reshaped = tf.reshape(input_char_windows, shape=(-1, self.window_size), name="reshape_input_char_windows")
		# shape = (?batch_size x max_windows, window_size)
		# char mask
		char_mask = Embedding(self.alphabet_size, self.num_of_classes, input_length=1, trainable=False, weights=[embedding_mask_weights], name="mask_embedding")(input_char_windows_reshaped[:, (self.window_size-1)//2])				
		char_mask = tf.reshape(char_mask,(-1, self.batch_max_windows, self.num_of_classes), name="reshape_char_mask")
		# Embedding layer
		x = Embedding(self.alphabet_size, self.embedding_size, input_length=self.window_size, trainable=True, name="sequence_embedding")(input_char_windows_reshaped)
		# x = (?batch_size, window_size, embedding_size)
		middle_char_embedding = x[:,(self.window_size-1)//2]

		# Convolution layers
		convolution_output = []
		for num_filters, filter_width in self.conv_layers:
			conv = Conv1D(filters=num_filters, kernel_size=filter_width, activation='tanh',
									name='Conv1D_{}_{}'.format(num_filters, filter_width))(x)
			# conv = (?batch_size, window_size-filter_size+1, num_filters)
			pool = GlobalMaxPooling1D(name='MaxPoolingOverTime_{}_{}'.format(num_filters, filter_width))(conv)
			# pool = (?batch_size, num_filters)
			convolution_output.append(pool)

		if convolution_output != []:
			x = Concatenate()(convolution_output)
			# x = (?batch_size, total_number_of_filters)
			x = Dropout(rate=self.cnn_dropout_rate)(x)

			# concatenate middle char
			x = Concatenate()([x, middle_char_embedding])

			self.total_number_of_filters = self.total_number_of_filters + self.embedding_size

		else:
			x = Flatten()(x)
			self.total_number_of_filters = self.window_size * self.embedding_size

		char_cnn_output = Dropout(rate=self.cnn_dropout_rate)(x)
		char_cnn_output = tf.reshape(char_cnn_output, shape=(-1, self.batch_max_windows, self.total_number_of_filters), name="char_cnn_output")
		# char_cnn_otput = (?batch_size, max_windows, total_filters)
		##########################################################################
		
		
		# get bert tokens coresponding to sent_ids and token_ids
		batch_indexes = tf.range(0, keras_internal_batch_size, name="range_batch_indexes")
		batch_indexes = tf.reshape(batch_indexes, (-1,1), name="reshape_batch_indexes")
		batch_indexes = tf.tile(batch_indexes, (1,self.batch_max_windows), name="tile_batch_indexes")
		indices = tf.stack([batch_indexes, input_sent_ids, input_token_ids], axis = 2)
		bert_tokens = tf.gather_nd(bert_output, indices, name="bert_tokens")
		# apply bert dropout here?
		# bert_tokens = (?batch_size, max_windows, bert_hidden_size)
		bert_cnn_concatenation = Concatenate()([bert_tokens, char_cnn_output])
		# bert_cnn_concatenation = char_cnn_output
		
		# hidden layer
		hidden = Dense(self.fc_hidden_size, activation='relu')(bert_cnn_concatenation)

		# Output layer
		predictions = Dense(self.num_of_classes, activation='softmax')(hidden)
		# mask predictions based on middle char 
		masked_predictions = keras.layers.multiply([predictions, char_mask])
		
		
		input_mask_reshaped = tf.reshape(input_mask, (-1, 1), name="reshape_input_mask")
		# mask prediction based on window mask
		# extended_mask = tf.reshape(input_mask, (-1, self.batch_max_windows, 1))
		# extended_mask = tf.tile(extended_mask, [1, 1, self.num_of_classes])	
		# masked_predictions = keras.layers.multiply([masked_predictions, extended_mask])
		
		flatten_masked_predictions = tf.reshape(masked_predictions, shape=(-1, self.num_of_classes), name="resh_flatmaskpred")
		# flatten_masked_predictions = masked_predictions
		# flatten_masked_prediction = (?batch_size x max_windows, num_of_classes)

		# Build and compile model
		model = Model(inputs=[input_bert_ids, input_bert_seg, input_token_ids, input_sent_ids, input_mask, input_char_windows], outputs=[flatten_masked_predictions, input_mask_reshaped])

		weights = np.ones(self.num_of_classes)
		model.compile(optimizer=self.optimizer, loss=[weighted_categorical_crossentropy(weights, self.num_of_classes).loss, None], metrics=[categorical_acc])
		# model.compile(optimizer=self.optimizer, loss=[self.loss, None], metrics=[tf.keras.metrics.categorical_accuracy])

		self.bert_wrapper.load_weights()
		if self.init_model != None:
			# TODO: make this automatic from main
			model.load_weights("rb/processings/diacritics/models/bert_models/" + self.init_model)
		self.model = model
		print("Bert+CharCNN model built: ")
		self.model.summary()
		

	def train(self, train_dataset, train_batch_size, train_size, dev_dataset, dev_batch_size, dev_size, epochs, file_evalname, char_to_id_dict, model_filename):

		best_wa_dia = -1
		best_wa_all = -1
		best_ca_dia = -1
		best_ca_all = -1
		best_epoch = -1

		dev_steps = (dev_size // dev_batch_size) + 1
		
		if dev_batch_size == 1:
			dev_steps += 1

		for i in range(epochs):
			print("EPOCH ", (i+1))
			self.model.fit(train_dataset, steps_per_epoch=train_size//train_batch_size, epochs=1, verbose=1)
			wa_dia, wa_all, ca_dia, ca_all, _ = utils.evaluate_model(self.model, file_evalname, dev_dataset, dev_steps)
			if wa_dia > best_wa_dia:
				best_wa_dia = wa_dia
				best_wa_all = wa_all
				best_ca_dia = ca_dia
				best_ca_all = ca_all
				best_epoch = i+1
				self.model.save(model_filename+".h5")
			
			print("Best model: epoch =", best_epoch, "best word_accuracy_dia =", format(best_wa_dia, '.4f'), "best word_accuracy_all =", format(best_wa_all, '.4f'), 
							"best char_accuracy_dia =", format(best_ca_dia, '.4f'), "best char_accuracy_all =", format(best_ca_all, '.4f'))
			print("---------------")

def categorical_acc(y_true, y_pred):
	# TODO: change this to number of classes
	y_true = tf.reshape(y_true, shape=(-1, 5), name="reshape_acc")
	return keras.metrics.categorical_accuracy(y_true, y_pred)


class weighted_categorical_crossentropy(object):
	"""
	A weighted version of keras.objectives.categorical_crossentropy

	Variables:
		weights: numpy array of shape (C,) where C is the number of classes

	Usage:
		loss = weighted_categorical_crossentropy(weights).loss
		model.compile(loss=loss,optimizer='adam')
	"""

	def __init__(self,weights,num_of_classes):
		self.weights = K.variable(weights)
		self.num_of_classes = num_of_classes

        
	def loss(self, y_true, y_pred):
		y_true = tf.reshape(y_true, shape=(-1, self.num_of_classes), name="reshape_loss")
		
		# scale preds so that the class probas of each sample sum to 1
		y_pred = y_pred / K.sum(y_pred, axis=-1, keepdims=True)
		
		# clip
		y_pred = K.clip(y_pred, K.epsilon(), 1)
			
		# calc
		loss = y_true*K.log(y_pred)*self.weights
		loss =-K.sum(loss,-1)
		return loss




from rb.core.lang import Lang
from rb.core.document import Document
from rb.core.text_element_type import TextElementType
from rb.complexity.complexity_index import ComplexityIndex, compute_indices
from rb.similarity.vector_model import VectorModelType, CorporaEnum, VectorModel
from rb.similarity.vector_model_factory import VECTOR_MODELS, create_vector_model
from rb.cna.cna_graph import CnaGraph
from typing import Tuple, List, Dict
import os
import numpy as np
import csv
import uuid
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class Feedback:


    def __init__(self):
        pass
    
    def get_vector_model(self, lang: Lang = Lang.RO) -> VectorModel:
        global logger
        if lang is Lang.RO:
            vector_model = create_vector_model(Lang.RO, VectorModelType.from_str('word2vec'), "readme")
        elif lang is Lang.EN:
            vector_model = create_vector_model(Lang.EN, VectorModelType.from_str("word2vec"), "coca")
        else:
            logger.error(f'Language {lang.value} is not supported for feedback task')
            vector_model = None
        return vector_model

    def get_used_indices(self) -> Dict[TextElementType, List[str]]:
        indices_files = ['ro_indices_word.txt', 'ro_indices_sent.txt', 'ro_indices_block.txt', 'ro_indices_doc.txt']
        indices_names = {TextElementType.DOC: [],
                         TextElementType.BLOCK: [],
                         TextElementType.SENT: [],
                         TextElementType.WORD: []}
        for i, in_file in enumerate(indices_files):
            lvl = None
            if i == 0:
                lvl = TextElementType.WORD
            elif i == 1:
                lvl = TextElementType.SENT
            elif i == 2:
                lvl = TextElementType.BLOCK
            else:
                lvl = TextElementType.DOC

            with open(os.path.join('rb/processings/readme_feedback', in_file), 'rt', encoding='utf-8') as f:
                for line in f:
                    if len(line.strip()) > 0:
                        indices_names[lvl].append(line.strip())
        return indices_names

    def compute_thresholds(self, values: List[float]) -> Tuple[int, int]:
        if len(values) > 1:
            stdev = np.std(values)
            avg = np.mean(values)
        elif len(values) == 1:
            avg = values[0]
            stdev = 1
        else:
            avg = -1
            stdev = -1
        return avg - 2.0 * stdev, avg + 2.0 * stdev 

    def compute_extreme_values(self,
        path_to_csv='categories_readme/en_stats.csv', output_file='readme_extreme_values.txt'):
        indices = self.get_used_indices()
        ind_name_to_row = {vv: 0 for v in list(indices.values()) for vv in v}
        ind_name_to_values = {vv: [] for v in list(indices.values()) for vv in v}
        ind_name_to_extreme_values = {vv: (0, 0) for v in list(indices.values()) for vv in v}

        stats = csv.reader(open(path_to_csv, 'rt', encoding='utf-8'))
        for i, row in enumerate(stats):
            if i == 0:
                for i, ind_name in enumerate(row):
                    if ind_name in ind_name_to_row:
                        ind_name_to_row[ind_name] = i
            else:
                for j, ind_value in enumerate(row):
                    for ind_name, ind_row in ind_name_to_row.items():
                        if ind_row == j:
                            ind_name_to_values[ind_name].append(float(ind_value))
                            break
        output_file = open(output_file, 'w')
        for ind_name, values in ind_name_to_values.items():
            ind_name_to_extreme_values[ind_name] = self.compute_thresholds(values)
            print(ind_name, ind_name_to_extreme_values[ind_name], file=output_file)
        return ind_name_to_extreme_values            

    def compute_indices_and_tokenization(self, text, lang: Lang) -> Dict[str, Dict[str, List]]:
        result = {
            TextElementType.WORD.name: {},
            TextElementType.SENT.name: {},
            TextElementType.BLOCK.name: {},
            TextElementType.DOC.name: {}
        }

        KEY_TEXT = "text"
        KEY_DOC = "doc"
        KEY_BLOCK = "block"
        KEY_SENT = "sentence"
        KEY_WORD = "word"
        KEY_VALUES = "values"

        indices = self.get_used_indices()
        doc = Document(lang=lang, text=text)
        vector_model = self.get_vector_model(lang=lang)
        cna_graph = CnaGraph(docs=doc, models=[vector_model])
        compute_indices(doc=doc, cna_graph=cna_graph)
        words, sents, blocks, docs = [], [], [], []
        ind_words, ind_sents, inds_blocks, inds_doc = indices[TextElementType.WORD],\
                  indices[TextElementType.SENT],  indices[TextElementType.BLOCK], indices[TextElementType.DOC]

        # doc
        d_ind_list = []
        for ind in indices[TextElementType.DOC]:
            for ind_name, ind_v in doc.indices.items():
                if repr(ind_name) == ind:
                    d_ind_list.append(ind_v)
        docs.append({KEY_TEXT: doc.text, KEY_BLOCK: 0, KEY_SENT: 0,
                             KEY_WORD: 0, KEY_VALUES: d_ind_list, KEY_DOC: 0})
        t_doc = []
        for i_block, block in enumerate(doc.get_blocks()):
            # block 
            b_ind_list = []
            for ind in indices[TextElementType.BLOCK]:
                for ind_name, ind_v in block.indices.items():
                    if repr(ind_name) == ind:
                        b_ind_list.append(ind_v)
            blocks.append({KEY_TEXT: block.text, KEY_BLOCK: i_block, KEY_SENT: 0,
                             KEY_WORD: 0, KEY_VALUES: b_ind_list, KEY_DOC: 0})

            t_block = []
            for i_sent, sent in enumerate(block.get_sentences()):
                # sent
                s_ind_list = []
                for ind in indices[TextElementType.SENT]:
                    for ind_name, ind_v in sent.indices.items():
                        if repr(ind_name) == ind:
                            s_ind_list.append(ind_v)
                sents.append({
                    KEY_TEXT: sent.text, KEY_BLOCK: i_block, KEY_SENT: i_sent, KEY_VALUES: s_ind_list,
                    KEY_WORD: 0, KEY_DOC: 0})

                t_sent = []
                for i_word, word in enumerate(sent.get_words()):
                    # word
                    w_ind_list = []
                    for ind in indices[TextElementType.WORD]:
                        for ind_name, ind_v in word.indices.items():
                            if repr(ind_name) == ind:
                                w_ind_list.append(ind_v)
                    words.append({KEY_TEXT: word.text, KEY_BLOCK: i_block, KEY_SENT: i_sent,
                             KEY_WORD: i_word, KEY_VALUES: w_ind_list, KEY_DOC: 0})
                    t_sent.append(word.text)
                t_block.append(t_sent)
            t_doc.append(t_block)

        result[TextElementType.DOC.name] = {
            'indices': indices[TextElementType.DOC],
            'results': docs
        }
        result[TextElementType.BLOCK.name] = {
            'indices': indices[TextElementType.BLOCK],
            'results': blocks
        }
        result[TextElementType.SENT.name] = {
            'indices': indices[TextElementType.SENT],
            'results': sents
        }
        result[TextElementType.WORD.name] = {
            'indices': indices[TextElementType.WORD],
            'results': words
        }

        result['TOKENIZATION'] = t_doc

        ff = open('debug.txt', 'w')
        print(result, file=ff)

        return result


from enum import Enum, auto

class DepEnum(Enum):
    ACL = auto()
    ADVCL = auto()
    ADVMOD = auto()
    AMOD = auto()
    APPOS = auto()
    AUX = auto()
    CASE = auto()
    CC = auto()
    CCOMP = auto()
    CLF = auto()
    COMPOUND = auto()
    CONJ = auto()
    COP = auto()
    CSUBJ = auto()
    DEP = auto()
    DET	 = auto()
    DISCOURSE = auto()
    DISLOCATED = auto()
    EXPL = auto()
    FIXED = auto()
    FLAT = auto()
    GOESWITH = auto()
    IOBJ = auto()
    LIST = auto()
    MARK = auto()
    NMOD = auto()
    NSUBJ = auto()
    NUMMOD = auto()
    OBJ = auto()
    OBL = auto()
    ORPHAN = auto()
    PARATAXIS = auto()
    PUNCT = auto()
    REPARANDUM = auto()
    ROOT = auto()
    VOCATIVE = auto()
    XCOMP = auto()



from typing import List, Tuple, Union

import spacy
from rb.core.lang import Lang
from rb.core.span import Span
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.core.word import Word
from rb.parser.spacy_parser import SpacyParser
from spacy.tokens import Doc

Dependency = Tuple[Word, Word, str]
Dependencies = List[Dependency]


class Sentence(TextElement):


    def __init__(self, lang: Lang, text: Union[spacy.tokens.Span, str],
                 depth: int = TextElementType.SENT.value,
                 container: TextElement = None):

        TextElement.__init__(self, lang=lang, text=text if isinstance(text, str) else text.text,
                             depth=depth, container=container)
        if isinstance(text, str):
            text = SpacyParser.get_instance().parse(text, lang)

        words = {token.i: Word(lang, token, container=self) for token in text}
        for word, token in zip(words.values(), text):
            word.head = words[token.head.i]
            if word.head is not word:
                word.head.children.append(word)
        self.entities = [Span(lang, text=ent.text, words=[words[token.i] for token in ent])
                         for ent in text.ents]
        if isinstance(text, Doc):
            self.root = words[[sent for sent in text.sents][0].root.i]
        else:
            self.root = words[text.root.i]
        self.components = [word for word in words.values()]

    def get_dependencies(self) -> Dependencies:
        return [(word.head, word, word.dep) 
                for word in self.components 
                if word.head != word]
            
    def get_sentences(self) -> List["Sentence"]:
        return [self]
    
    def get_words(self) -> List[Word]:
        return self.components


from rb.processings.pipeline.estimator import Regressor
from rb.processings.pipeline.dataset import Dataset, Task
from typing import List, Dict
from sklearn import svm

class SVR(Regressor):
    def __init__(self, dataset: Dataset, tasks: List[Task], params: Dict[str, str]):
        super().__init__(dataset, tasks, params)
        self.model = svm.SVR(kernel=params["kernel"])

    @classmethod
    def parameters(cls):
        return {
            "kernel": ["rbf", "poly", "sigmoid"]
        }


from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType   
from typing import List, Callable
import pyphen

from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class WdSyllab(ComplexityIndex):

    
    def __init__(self, lang: Lang,
            reduce_depth: int, reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.WORD,
                                 abbr="WdSyll", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)
        if lang is lang.RO:
            self.pyphen = pyphen.Pyphen(lang='ro')
        elif lang is lang.EN:
            self.pyphen = pyphen.Pyphen(lang='en')
        elif lang is lang.ES:
            self.pyphen = pyphen.Pyphen(lang='es')
        elif lang is lang.FR:
            self.pyphen = pyphen.Pyphen(lang='fr')
        elif lang is lang.DE:
             self.pyphen = pyphen.Pyphen(lang='de')
        elif lang is lang.IT:
             self.pyphen = pyphen.Pyphen(lang='it')
        elif lang is lang.NL:
             self.pyphen = pyphen.Pyphen(lang='nl')
        elif lang is lang.RU:
             self.pyphen = pyphen.Pyphen(lang='ru')
    
    def process(self, element: TextElement) -> float:
        return self.reduce_function(self.compute_above(element))

    def compute_below(self, element: TextElement) -> float:
        if element.is_word() == True:
            return len(self.pyphen.inserted(element.text).split('-'))
        elif element.depth <= self.reduce_depth:
            res = 0
            for child in element.components:
                res += self.compute_below(child)
            return res
    
    def compute_above(self, element: TextElement) -> List[float]:
        if element.depth > self.reduce_depth:
            values = []
            for child in element.components:
                values += self.compute_above(child)
            element.indices[self] = self.reduce_function(values)
        elif element.depth == self.reduce_depth:
            values = [self.compute_below(element)]
            element.indices[self] = self.reduce_function(values)

        else:
            logger.error('wrong reduce depth value.')
        return values
    
    def __getstate__(self):
        return {
            key: value
            for key, value in self.__dict__.items()
            if key != "pyphen"
        }


import statistics
import random
import sys
import pickle
import tensorflow as tf
import os
from sklearn.utils import class_weight
import numpy as np
from collections import Counter
from bert.tokenization.bert_tokenization import FullTokenizer

# split train file in train and dev
def splitDataset(filepath, train_ratio):
    
    full_sentences = []
    with open(filepath, 'r', encoding='utf-8') as in_file:
        for sentence in in_file:
            full_sentences.append(sentence)


    full_sentences = set(full_sentences)
    sentences = len(full_sentences)
    print("Total sentences =", format(sentences, ",d"))

    dev_count = int((1-train_ratio) * sentences)
    print("Dev sentences =", format(dev_count, ",d"))

    # pick random sentences to be in dev
    random_indexes = random.sample(range(sentences), k=dev_count)
    random_indexes.append(9999999999999)
    random_indexes.sort()


    dev_sentences = []
    train_sentences = []

    random_index = 0
    for sentence_index, sentence in enumerate(full_sentences):
        if sentence_index < random_indexes[random_index]:
            train_sentences.append(sentence)

        elif sentence_index == random_indexes[random_index]:
            random_index += 1
            dev_sentences.append(sentence)
        else:
            print("SHOULD NOT HAPPEN")
            sys.exit()

    print("dev =", format(len(dev_sentences), ",d"), "train =", format(len(train_sentences), ",d"), "total =", format(len(dev_sentences)+len(train_sentences), ",d"))
    inter = list(set(dev_sentences) & set(train_sentences)) 
    print("Intersection =", len(inter))
    
    with open(filepath.split(".")[0]+"_tr.txt", 'w', encoding='utf-8') as train_file:
        train_file.writelines(train_sentences)
    with open(filepath.split(".")[0]+"_de.txt", 'w', encoding='utf-8') as dev_file:
        dev_file.writelines(dev_sentences)

# build char vocab
def build_char_vocab():
    dict_train = statistics.compute_chars_dict("dataset/split/train.txt")
    dict_dev = statistics.compute_chars_dict("dataset/split/dev.txt")
    dict_test = statistics.compute_chars_dict("dataset/split/test.txt")

    full_counter = dict_train + dict_dev + dict_test
    print(full_counter)
    print(len(full_counter))


    char_dict = {}
    char_dict["<PAD>"] = 0
    for char, _ in full_counter.items():
        char_dict[char] = len(char_dict)

    print(char_dict)
    print(len(char_dict))

    pickle.dump(char_dict, open("dataset/split/char_dict", "wb"))

# remove char diacritic
# a -> a
# ă -> a etc
def get_char_basic(char):

    if char == 'a' or char == 'ă' or char == 'â':
        return 'a'
    
    if char == "s" or char == "ș":
        return 's'
    
    if char == "t" or char == "ț":
        return 't'

    if char == "i" or char == "î":
        return 'i'
    
    return char

# get label of transformation from char w/o diacritic to char with diacritic
# no_change -> 0
# a to ă -> 1
# a/i to â/î -> 2
# s to ș -> 3
# t to ț -> 4
def get_label(basic_char, diacritic_char):

    if basic_char == diacritic_char:
        return 0

    if basic_char == 'a':
        if diacritic_char == 'ă':
            return 1
        elif diacritic_char == 'â':
            return 2
        else:
            print("Wrong utils.get_label", basic_char, "->", diacritic_char)

    if basic_char == 'i':
        if diacritic_char == "î":
            return 2
        else:
            print("Wrong utils.get_label", basic_char, "->", diacritic_char)

    if basic_char == 's':
        if diacritic_char == "ș":
            return 3
        else:
            print("Wrong utils.get_label", basic_char, "->", diacritic_char)

    if basic_char == 't':
        if diacritic_char == "ț":
            return 4
        else:
            print("Wrong utils.get_label", basic_char, "->", diacritic_char)

# get predicted char from basic char and predicted class
def get_char_from_label(basic_char, predicted_class):

    if predicted_class == 0:
        return basic_char
    
    if basic_char == 'a' and predicted_class == 1:
        return "ă"
    if basic_char == 'a' and predicted_class == 2:
        return "â"
    
    if basic_char == 'i' and predicted_class == 2:
        return "î"

    if basic_char == 's' and predicted_class == 3:
        return "ș"

    if basic_char == 't' and predicted_class == 4:
        return "ț"
    
    else:
        print("Error in utils.get_char_from_label")
        print("basic_char =", basic_char, "predicted_class =", predicted_class)
        sys.exit()

# generator features for cnn: only window and label
def generator_cnn_features(filepath, char_to_id_dict, window_size):
    diacritics = set("aăâiîsștț")
    id_to_char_dict = {v: k for k, v in char_to_id_dict.items()}

    with open(filepath, "r", encoding='utf-8') as in_file:
        for _, sentence in enumerate(in_file):
            
            char_ids = [char_to_id_dict[get_char_basic(char)] for char in sentence]
            values_to_pad = (window_size-1)//2
            char_ids = [char_to_id_dict["<PAD>"]] * values_to_pad + char_ids + [char_to_id_dict["<PAD>"]] * values_to_pad
            
            for char_index, char in enumerate(sentence):
                # we padded vector
                char_index += values_to_pad
                if char in diacritics:
                    full_window = char_ids[(char_index - values_to_pad):(char_index + values_to_pad + 1)]
                    label = get_label(id_to_char_dict[char_ids[char_index]], char)
                    
                    categorical = np.zeros((5))
                    categorical[label] = 1
                    yield np.array(full_window), categorical

# generator features for bert+cnn: bert_tokens, window, position (relative to bert_tokens) and label
# this generator works at sentence level
# for each sentence it returns -> bert_inputs + list of token_idx + list of windows + list of labels
# bert_input_ids :: [bert_max_seq_len]
# bert_segment_ids :: [bert_max_seq_len]
# token_ids :: [no_windows] this is not fixed, it depends on the sentence
# char_windows :: [no_windows, window_size] # this is actually list of numpy arrays
# labels :: [no_windows, no_classes] 
def generator_sentence_bert_cnn_features(filepath, char_to_id_dict, window_size, bert_wrapper):
    
    diacritics = "aăâiîsștț"
    id_to_char_dict = {v: k for k, v in char_to_id_dict.items()}

    with open(filepath, "r", encoding='utf-8') as in_file:
        for _, sentence in enumerate(in_file):

            basic_sentence = ''.join([get_char_basic(char) for char in sentence])
            tokens = bert_wrapper.tokenizer.tokenize(basic_sentence)
            sentence_bert_input_ids, sentence_bert_segment_ids = bert_wrapper.process_text(basic_sentence)
            sentence_token_ids = []
            sentence_char_cnn_windows = []
            sentence_labels = []
            # save map: char_diac_index -> bert_tokenizer_index
            # example: privata -> 'privat' + '##a'
            # 0 (first a) -> 0
            # 1 (second a) -> 1
            sentence_char_index = {}
            
            char_index = 0
            for token_index, token in enumerate(tokens):
                for char in token:
                    if char in diacritics:
                        # offset by 1 because of '[CLS]'
                        sentence_char_index[char_index] = token_index + 1
                        char_index += 1
            
            char_ids = [char_to_id_dict[get_char_basic(char)] for char in sentence]
            values_to_pad = (window_size-1)//2
            char_ids = [char_to_id_dict["<PAD>"]] * values_to_pad + char_ids + [char_to_id_dict["<PAD>"]] * values_to_pad

            char_dia_index = 0
            for char_index, char in enumerate(sentence):
                # we padded vector
                char_index += values_to_pad
                if char in diacritics:

                    full_window = char_ids[(char_index - values_to_pad):(char_index + values_to_pad + 1)]
                    label = get_label(id_to_char_dict[char_ids[char_index]], char)
                    
                    categorical = np.zeros((5))
                    categorical[label] = 1
                    
                    sentence_labels.append(categorical)
                    sentence_token_ids.append(sentence_char_index[char_dia_index])
                    sentence_char_cnn_windows.append(full_window)
                    
                    char_dia_index += 1

                    # print(sentence_bert_input_ids, sentence_bert_segment_ids[char_dia_index], sentence_token_ids, sentence_char_cnn_windows,sentence_labels)
            # sys.exit()
            yield sentence_bert_input_ids, sentence_bert_segment_ids, sentence_token_ids, sentence_char_cnn_windows, sentence_labels
    

# high level generator for bert+cnn   
# output
# bert_input_ids :: [max_sentences, bert_max_seq_len]
# bert_segment_ids :: [max_sentences, bert_max_seq_len]
# token_ids :: [max_windows] for referencing token  
# sent_ids :: [max_windows] for referencing sentence
# mask :: [max_windows] for marking if a window is part of dataset or padding 
# char_windows :: [max_windows, window_size]
# labels :: [max_windows, no_classes]
def generator_bert_cnn_features(filepath, char_to_id_dict, window_size, bert_wrapper, max_sentences, max_windows,):

    padding_window = [0] * window_size
    padding_labels = np.array([0, 0, 0, 0, 0])
    padding_input_ids, padding_segment_ids = bert_wrapper.process_text("")

    crt_sentences = 0
    crt_windows = 0

    bert_input_ids = []
    bert_segment_ids =[]
    token_ids = []
    sentence_ids = []
    windows_mask = []
    char_windows = []
    labels = []

    sentence_generator = generator_sentence_bert_cnn_features(filepath, char_to_id_dict, window_size, bert_wrapper)
    for sentence_entry in sentence_generator:
        sentence_bert_input_ids, sentence_bert_segment_ids, sentence_token_ids, sentence_char_cnn_windows, sentence_labels = sentence_entry
        # print(sentence_bert_input_ids, sentence_bert_segment_ids, sentence_token_ids, sentence_char_cnn_windows, sentence_labels)
        
        bert_input_ids.append(sentence_bert_input_ids)
        bert_segment_ids.append(sentence_bert_segment_ids)

        for window_index in range(len(sentence_token_ids)):
            token_ids.append(sentence_token_ids[window_index])
            sentence_ids.append(crt_sentences)
            windows_mask.append(1.0)
            char_windows.append(sentence_char_cnn_windows[window_index])
            labels.append(sentence_labels[window_index])
            
            crt_windows += 1

            if crt_windows == max_windows:

                sentences_to_pad = max_sentences - crt_sentences - 1
                bert_input_ids = bert_input_ids + [padding_input_ids] * sentences_to_pad
                bert_segment_ids = bert_segment_ids + [padding_segment_ids] * sentences_to_pad
                
                yield {'bert_input_ids':bert_input_ids, 'bert_segment_ids':bert_segment_ids, 'token_ids': token_ids, 
                    'sent_ids': sentence_ids, 'mask': windows_mask, 'char_windows': char_windows}, labels

                # take the last sentence before padding
                bert_input_ids = [bert_input_ids[crt_sentences]]
                bert_segment_ids = [bert_segment_ids[crt_sentences]]
                # reset global vars
                crt_sentences = 0
                crt_windows = 0

                token_ids = []
                sentence_ids = []
                windows_mask = []
                char_windows = []
                labels = []

        crt_sentences += 1
        if crt_sentences == max_sentences:
            # we have reached maximum sentence count
            # we need to pad up to max_window_size
            values_to_pad = max_windows - crt_windows

            token_ids = token_ids + [0] * values_to_pad
            sentence_ids = sentence_ids + [0] * values_to_pad
            windows_mask = windows_mask + [0] * values_to_pad
            char_windows = char_windows + [padding_window] * values_to_pad
            labels = labels + [padding_labels] * values_to_pad

            yield {'bert_input_ids':bert_input_ids, 'bert_segment_ids':bert_segment_ids, 'token_ids': token_ids, 
                    'sent_ids': sentence_ids, 'mask': windows_mask, 'char_windows': char_windows}, labels

            # reset global vars
            crt_sentences = 0
            crt_windows = 0
            bert_input_ids = []
            bert_segment_ids =[]
            token_ids = []
            sentence_ids = []
            windows_mask = []
            char_windows = []
            labels = []

    # return uncompleted
    # we have to pad up to max_sentences and max_windows
    # pad up to max_sentences
    sentences_to_pad = max_sentences - crt_sentences
    bert_input_ids = bert_input_ids + [padding_input_ids] * sentences_to_pad
    bert_segment_ids = bert_segment_ids + [padding_segment_ids] * sentences_to_pad

    # pad up to max_windows
    values_to_pad = max_windows - crt_windows
    token_ids = token_ids + [0] * values_to_pad
    sentence_ids = sentence_ids + [0] * values_to_pad
    windows_mask = windows_mask + [0] * values_to_pad
    char_windows = char_windows + [padding_window] * values_to_pad
    labels = labels + [np.zeros(5)] * values_to_pad
    
    # print("BII", len(bert_input_ids))#, bert_input_ids)
    # print("BSI", len(bert_segment_ids))#, bert_segment_ids)
    # print("Token ids", len(token_ids))#, token_ids)
    # print("Sent ids", len(sentence_ids))#, sentence_ids)
    # print("Window mask", len(windows_mask))#, windows_mask)
    # print("Char windows", len(char_windows))#, char_windows)
    # print("Labels", len(labels))#, labels)

    yield {'bert_input_ids':bert_input_ids, 'bert_segment_ids':bert_segment_ids, 'token_ids': token_ids, 
            'sent_ids': sentence_ids, 'mask': windows_mask, 'char_windows': char_windows}, labels

# from diacritics site
# word level accuracy on word that accept dia 
def compute_word_accuracy_dia(true_words, pred_words):
    is_valid = lambda word: any(c in word for c in 'aăâiîsștț')
    n_correct = sum(t == p for t, p in zip(true_words, pred_words) if is_valid(t))
    n_total = sum(is_valid(t) for t in true_words)
    return n_correct / n_total

# word level accuracy on all words
def compute_word_accuracy(true_words, pred_words):
    is_valid = lambda word: True
    n_correct = sum(t == p for t, p in zip(true_words, pred_words) if is_valid(t))
    n_total = sum(is_valid(t) for t in true_words)
    return n_correct / n_total

# word level accuracy on char that accept dia 
def compute_char_accuracy_dia(true_chars, pred_chars):
    is_valid = lambda char: char in 'aăâiîsștț'
    n_correct = sum(t == p for t, p in zip(true_chars, pred_chars) if is_valid(t))
    n_total = sum(is_valid(t) for t in true_chars)
    return n_correct / n_total

# char level accuracy on all chars
def compute_char_accuracy(true_chars, pred_chars):
    is_valid = lambda word: True
    n_correct = sum(t == p for t, p in zip(true_chars, pred_chars) if is_valid(t))
    n_total = sum(is_valid(t) for t in true_chars)
    return n_correct / n_total


def evaluate_model_on_file(model, filepath, char_to_id_dict, window_size):
        
    diacritics = "aăâiîsștț"

    global_true_words = []
    global_predicted_words = []

    global_true_chars = []
    global_predicted_chars = []

    predicted_chars = []

    predicted_cla = []
    predicted_dia = []

    with open(filepath, "r", encoding='utf-8') as in_file:
        sentence_windows = []
        predicted_sentence = []
        predicted_indexes = []
        basic_chars = []
        for sentence_index, sentence in enumerate(in_file):

            global_true_chars.extend(sentence)
            sentence_true_words = sentence.split(" ")
            global_true_words.extend(sentence_true_words)

            # bring chars to base form
            basic_sentence = ''.join(list(map(lambda x: get_char_basic(x), sentence)))
            char_ids = list(map(lambda x: char_to_id_dict[x], basic_sentence))

            values_to_pad = (window_size-1)//2
            for _ in range(values_to_pad):
                char_ids.append(char_to_id_dict["<PAD>"])
                char_ids.insert(0, char_to_id_dict["<PAD>"])

            for char_index, char in enumerate(sentence):
                char_index += values_to_pad
                if char not in diacritics:
                    predicted_sentence.append(char)
                    predicted_chars.append(char)
                else:
                    # generate window
                    left = char_ids[char_index-values_to_pad:char_index]
                    right = char_ids[char_index+1:char_index+values_to_pad+1]
                    full_window = left + [char_to_id_dict[get_char_basic(char)]] + right
                    
                    sentence_windows.append(full_window)

                    predicted_indexes.append(len(predicted_sentence))
                    predicted_sentence.append("X")
                    predicted_chars.append("X")
                    basic_chars.append(get_char_basic(char))
            

            if sentence_index % 1e4 == 0 and sentence_index != 0 :

                prediction_vectors = model.predict(sentence_windows)
                for index, prediction_vector in enumerate(prediction_vectors):
                    predicted_class = np.argmax(prediction_vector)
                    
                    predicted_char = get_char_from_label(basic_chars[index], predicted_class)
                    predicted_chars[predicted_indexes[index]] = predicted_char
                    
                    predicted_dia.append(predicted_char)
                    predicted_cla.append(predicted_class)

                predicted_sentence = ''.join(predicted_chars).replace("\n", "\n ").split(" ")[:-1]

                global_predicted_words.extend(predicted_sentence)
                global_predicted_chars.extend(predicted_chars)

                # print(sentence_index, len(sentence_windows))
                # print(global_true_words)
                # print(global_predicted_words)
                
                sentence_windows = []
                predicted_sentence = []
                predicted_indexes = []
                predicted_chars = []
                basic_chars = []


    if sentence_windows != []:
        prediction_vectors = model.predict(sentence_windows)
        for index, prediction_vector in enumerate(prediction_vectors):
            predicted_class = np.argmax(prediction_vector)
            
            predicted_char = get_char_from_label(basic_chars[index], predicted_class)
            predicted_chars[predicted_indexes[index]] = predicted_char
            
            predicted_dia.append(predicted_char)
            predicted_cla.append(predicted_class)

        predicted_sentence = ''.join(predicted_chars).replace("\n", "\n ").split(" ")[:-1]

        global_predicted_words.extend(predicted_sentence)
        global_predicted_chars.extend(predicted_chars)



    if len(global_true_words) != len(global_predicted_words):
        print("Mismatch between #true_words and #predicted_words")
        print(len(global_true_words), len(global_predicted_words))
        sys.exit()


    if len(global_predicted_chars) != len(global_predicted_chars):
        print("Mismatch between #true_chars and #predicted_chars")
        print(len(global_true_chars), len(global_predicted_chars))
        sys.exit()
    
    word_accuracy_dia = compute_word_accuracy_dia(global_true_words, global_predicted_words)
    word_accuracy = compute_word_accuracy(global_true_words, global_predicted_words)

    char_accuracy_dia = compute_char_accuracy_dia(global_true_chars, global_predicted_chars)
    char_accuracy = compute_char_accuracy(global_true_chars, global_predicted_chars)


    print("Word accuracy dia =", format(word_accuracy_dia, '.4f'))
    print("Word accuracy all =", format(word_accuracy, '.4f'))

    print("Char accuracy dia =", format(char_accuracy_dia, '.4f'))
    print("Char accuracy all =", format(char_accuracy, '.4f'))

    # print(len(predicted_dia), len(predicted_cla))
    print(Counter(predicted_dia), Counter(predicted_cla))

    return word_accuracy_dia, word_accuracy, char_accuracy_dia, char_accuracy, global_predicted_words


def evaluate_model(model, filepath, dataset, steps, model_type="BertCNN", write_to_file=False, outfile_name=None):

    diacritics = set("aăâiîsștț")
    predictions = model.predict(dataset, steps=steps)
    
    if model_type == "BertCNN":
        filtered_predictions = []
        for index in range(len(predictions[0])):
            if predictions[1][index] == 1:
                filtered_predictions.append(predictions[0][index])
            
        predictions = np.array(filtered_predictions)
    
    predicted_classes = list(map(lambda x: np.argmax(x), predictions))
    print(predictions.shape, len(predicted_classes))
    predicted_dia = []
    predicted_cla = []

    global_true_words = []
    global_predicted_words = []

    global_true_chars = []
    global_predicted_chars = []

    prediction_index = 0

    with open(filepath, "r", encoding='utf-8') as in_file:
        for _, sentence in enumerate(in_file):

            global_true_chars.extend(sentence)
            sentence_true_words = sentence.split(" ")
            global_true_words.extend(sentence_true_words)

            for _, char in enumerate(sentence):
                if char in diacritics:
                    basic_char = get_char_basic(char)
                    predicted_char = get_char_from_label(basic_char, predicted_classes[prediction_index])
                    global_predicted_chars.append(predicted_char)
        
                    predicted_dia.append(predicted_char)
                    predicted_cla.append(predicted_classes[prediction_index])

                    prediction_index += 1

                else: 
                    global_predicted_chars.append(char)

    global_predicted_words = ''.join(global_predicted_chars).replace("\n", "\n ").split(" ")[:-1]

    if len(global_true_words) != len(global_predicted_words):
        print("Mismatch between #true_words and #predicted_words")
        print(len(global_true_words), len(global_predicted_words))
        sys.exit()


    if len(global_predicted_chars) != len(global_predicted_chars):
        print("Mismatch between #true_chars and #predicted_chars")
        print(len(global_true_chars), len(global_predicted_chars))
        sys.exit()
    
    word_accuracy_dia = compute_word_accuracy_dia(global_true_words, global_predicted_words)
    word_accuracy = compute_word_accuracy(global_true_words, global_predicted_words)

    char_accuracy_dia = compute_char_accuracy_dia(global_true_chars, global_predicted_chars)
    char_accuracy = compute_char_accuracy(global_true_chars, global_predicted_chars)


    print("Word accuracy dia =", format(word_accuracy_dia, '.4f'))
    print("Word accuracy all =", format(word_accuracy, '.4f'))

    print("Char accuracy dia =", format(char_accuracy_dia, '.4f'))
    print("Char accuracy all =", format(char_accuracy, '.4f'))

    # print(len(predicted_dia), len(predicted_cla))
    print(Counter(predicted_dia), Counter(predicted_cla))

    if write_to_file == True:
        # also write to file
        with open(outfile_name , "w", encoding="utf-8") as outfile:
            for word in global_predicted_words:
                if word[-1] == "\n":
                    outfile.write(word)
                else:
                    outfile.write(word + " ")

    return word_accuracy_dia, word_accuracy, char_accuracy_dia, char_accuracy, global_predicted_words


def generator_sentence_bert_cnn_features_string(string, char_to_id_dict, window_size, bert_wrapper):
    
    diacritics = "aăâiîsștț"
    id_to_char_dict = {v: k for k, v in char_to_id_dict.items()}

    sentence = string

    basic_sentence = ''.join([get_char_basic(char) for char in sentence])
    tokens = bert_wrapper.tokenizer.tokenize(basic_sentence)
    sentence_bert_input_ids, sentence_bert_segment_ids = bert_wrapper.process_text(basic_sentence)
    sentence_token_ids = []
    sentence_char_cnn_windows = []
    sentence_labels = []
    # save map: char_diac_index -> bert_tokenizer_index
    # example: privata -> 'privat' + '##a'
    # 0 (first a) -> 0
    # 1 (second a) -> 1
    sentence_char_index = {}
    
    char_index = 0
    for token_index, token in enumerate(tokens):
        for char in token:
            if char in diacritics:
                # offset by 1 because of '[CLS]'
                sentence_char_index[char_index] = token_index + 1
                char_index += 1
    
    char_ids = [char_to_id_dict[get_char_basic(char)] for char in sentence]
    values_to_pad = (window_size-1)//2
    char_ids = [char_to_id_dict["<PAD>"]] * values_to_pad + char_ids + [char_to_id_dict["<PAD>"]] * values_to_pad

    char_dia_index = 0
    for char_index, char in enumerate(sentence):
        # we padded vector
        char_index += values_to_pad
        if char in diacritics:

            full_window = char_ids[(char_index - values_to_pad):(char_index + values_to_pad + 1)]
            label = get_label(id_to_char_dict[char_ids[char_index]], char)
            
            categorical = np.zeros((5))
            categorical[label] = 1
            
            sentence_labels.append(categorical)
            sentence_token_ids.append(sentence_char_index[char_dia_index])
            sentence_char_cnn_windows.append(full_window)
            
            char_dia_index += 1

            # print(sentence_bert_input_ids, sentence_bert_segment_ids[char_dia_index], sentence_token_ids, sentence_char_cnn_windows,sentence_labels)
    # sys.exit()
    yield sentence_bert_input_ids, sentence_bert_segment_ids, sentence_token_ids, sentence_char_cnn_windows, sentence_labels


def generator_bert_cnn_features_string(string, char_to_id_dict, window_size, bert_wrapper, max_sentences, max_windows):

    padding_window = [0] * window_size
    padding_labels = np.array([0, 0, 0, 0, 0])
    padding_input_ids, padding_segment_ids = bert_wrapper.process_text("")

    crt_sentences = 0
    crt_windows = 0

    bert_input_ids = []
    bert_segment_ids =[]
    token_ids = []
    sentence_ids = []
    windows_mask = []
    char_windows = []
    labels = []

    sentence_generator = generator_sentence_bert_cnn_features_string(string, char_to_id_dict, window_size, bert_wrapper)
    for sentence_entry in sentence_generator:
        sentence_bert_input_ids, sentence_bert_segment_ids, sentence_token_ids, sentence_char_cnn_windows, sentence_labels = sentence_entry
        # print(sentence_bert_input_ids, sentence_bert_segment_ids, sentence_token_ids, sentence_char_cnn_windows, sentence_labels)
        
        bert_input_ids.append(sentence_bert_input_ids)
        bert_segment_ids.append(sentence_bert_segment_ids)

        for window_index in range(len(sentence_token_ids)):
            token_ids.append(sentence_token_ids[window_index])
            sentence_ids.append(crt_sentences)
            windows_mask.append(1.0)
            char_windows.append(sentence_char_cnn_windows[window_index])
            labels.append(sentence_labels[window_index])
            
            crt_windows += 1

            if crt_windows == max_windows:

                sentences_to_pad = max_sentences - crt_sentences - 1
                bert_input_ids = bert_input_ids + [padding_input_ids] * sentences_to_pad
                bert_segment_ids = bert_segment_ids + [padding_segment_ids] * sentences_to_pad
                
                yield {'bert_input_ids':bert_input_ids, 'bert_segment_ids':bert_segment_ids, 'token_ids': token_ids, 
                    'sent_ids': sentence_ids, 'mask': windows_mask, 'char_windows': char_windows}, labels

                # take the last sentence before padding
                bert_input_ids = [bert_input_ids[crt_sentences]]
                bert_segment_ids = [bert_segment_ids[crt_sentences]]
                # reset global vars
                crt_sentences = 0
                crt_windows = 0

                token_ids = []
                sentence_ids = []
                windows_mask = []
                char_windows = []
                labels = []

        crt_sentences += 1
        if crt_sentences == max_sentences:
            # we have reached maximum sentence count
            # we need to pad up to max_window_size
            values_to_pad = max_windows - crt_windows

            token_ids = token_ids + [0] * values_to_pad
            sentence_ids = sentence_ids + [0] * values_to_pad
            windows_mask = windows_mask + [0] * values_to_pad
            char_windows = char_windows + [padding_window] * values_to_pad
            labels = labels + [padding_labels] * values_to_pad

            yield {'bert_input_ids':bert_input_ids, 'bert_segment_ids':bert_segment_ids, 'token_ids': token_ids, 
                    'sent_ids': sentence_ids, 'mask': windows_mask, 'char_windows': char_windows}, labels

            # reset global vars
            crt_sentences = 0
            crt_windows = 0
            bert_input_ids = []
            bert_segment_ids =[]
            token_ids = []
            sentence_ids = []
            windows_mask = []
            char_windows = []
            labels = []

    # return uncompleted
    # we have to pad up to max_sentences and max_windows
    # pad up to max_sentences
    sentences_to_pad = max_sentences - crt_sentences
    bert_input_ids = bert_input_ids + [padding_input_ids] * sentences_to_pad
    bert_segment_ids = bert_segment_ids + [padding_segment_ids] * sentences_to_pad

    # pad up to max_windows
    values_to_pad = max_windows - crt_windows
    token_ids = token_ids + [0] * values_to_pad
    sentence_ids = sentence_ids + [0] * values_to_pad
    windows_mask = windows_mask + [0] * values_to_pad
    char_windows = char_windows + [padding_window] * values_to_pad
    labels = labels + [np.zeros(5)] * values_to_pad
    
    # print("BII", len(bert_input_ids))#, bert_input_ids)
    # print("BSI", len(bert_segment_ids))#, bert_segment_ids)
    # print("Token ids", len(token_ids))#, token_ids)
    # print("Sent ids", len(sentence_ids))#, sentence_ids)
    # print("Window mask", len(windows_mask))#, windows_mask)
    # print("Char windows", len(char_windows))#, char_windows)
    # print("Labels", len(labels))#, labels)

    yield {'bert_input_ids':bert_input_ids, 'bert_segment_ids':bert_segment_ids, 'token_ids': token_ids, 
            'sent_ids': sentence_ids, 'mask': windows_mask, 'char_windows': char_windows}, labels








if __name__ == "__main__":
    print("utils.py")
    # build_char_vocab()




from typing import Dict, List

from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.core.cscl.cscl_indices import CsclIndices
from rb.utils.rblogger import Logger


class Participant:

	def __init__(self, participant_id: str):
		self.participant_id = participant_id
		self.indices = dict()

		self.eligible_contributions = []
		self.significant_contributions = []

	def get_id(self) -> str:
		return self.participant_id

	def get_index(self, index: str):
		if index in self.indices:
			return self.indices[index]

		return 0

	def set_index(self, index: str, value: float):
		self.indices[index] = value

	def add_eligible_contribution(self, contribution):
		self.eligible_contributions.append(contribution)

	def add_significant_contribution(self, contribution):
		self.significant_contributions.append(contribution)



from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType   
from rb.core.sentence import Sentence
from typing import List, Callable
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class NoRepetitions(ComplexityIndex):

    
    def __init__(self, lang: Lang, window_size: int, 
        reduce_depth: int, reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.WORD,
                                 abbr="Repetitions", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)
        self.window_size = window_size

    def process(self, element: TextElement) -> float:
        return self.reduce_function(self.compute_above(element))

    def compute_repetitions(self, sent: Sentence):
        # TODO count also synonyms, not just same lemmas
        count_reps = 0

        for start in range(max(1, len(sent.components) - self.window_size)):
            word1 = sent.components[start]
            for i in range(min(len(sent.components) - 1, start + 1), min(len(sent.components), start + self.window_size)):
                word2 = sent.components[i]
                if word1.lemma == word2.lemma:
                    count_reps += 1

        return count_reps


    def compute_below(self, element: TextElement) -> List[float]:
        if element.is_sentence() == True:
            res = self.compute_repetitions(element)
            return [res]
        elif element.depth <= self.reduce_depth:
            res = []
            for child in element.components:
                res += self.compute_below(child)
            return res
    
    def compute_above(self, element: TextElement) -> List[float]:
        if element.depth > self.reduce_depth:
            values = []
            for child in element.components:
                values += self.compute_above(child)
            element.indices[self] = self.reduce_function(values)
        elif element.depth == self.reduce_depth:
            values = self.compute_below(element)
            element.indices[self] = self.reduce_function(values)
        else:
            logger.error('wrong reduce depth value.')
        return values


from enum import Enum, unique

@unique
class CsclIndices(Enum):
    SCORE = 'scor'
    SOCIAL_KB = 'social_kb'
    NO_CONTRIBUTION = 'no_contribution'
    NO_NOUNS = 'no_nouns'
    NO_VERBS = 'no_verbs'
    OUTDEGREE = 'outdegree'
    INDEGREE = 'indegree'
    NO_NEW_THREADS = 'no_new_threads'
    NEW_THREADS_OVERALL_SCORE = 'new_threads_overall_score'
    NEW_THREADS_CUMULATIVE_SOCIAL_KB = 'new_threads_cumulative_social_kb'
    NEW_THREADS_INTER_ANIMATION_DEGREE = 'new_threads_inter_animation_degree'
    AVERAGE_LENGTH_NEW_THREADS = 'average_length_new_threads'


from typing import Callable, Dict, List, Tuple, Union

import networkx as nx
import numpy as np
from rb.cna.edge_type import EdgeType
from rb.cna.overlap_type import OverlapType
from rb.core.block import Block
from rb.core.cscl.community import Community
from rb.core.cscl.contribution import Contribution
from rb.core.cscl.conversation import Conversation
from rb.core.document import Document
from rb.core.lang import Lang
from rb.core.meta_document import MetaDocument
from rb.core.pos import POS
from rb.core.text_element import TextElement
from rb.core.word import Word
from rb.similarity.vector_model import VectorModel


class CnaGraph:
	def __init__(self, docs: Union[TextElement, List[TextElement]], models: List[VectorModel]):
		if isinstance(docs, TextElement):
			docs = [docs]
		self.graph = nx.MultiDiGraph()
		self.models = models
		if all(isinstance(doc, Community) or isinstance(doc, MetaDocument) for doc in docs):
			for doc in docs:
				self.graph.add_node(doc)
				for conv in doc.components:
					nodes = self.add_element(conv)
					self.graph.add_edge(conv, doc, type=EdgeType.PART_OF)
					self.add_links(nodes)
		else:
			for doc in docs:
				self.add_element(doc)
			self.add_links(self.graph.nodes)
		
		self.importance = self.compute_importance()

		self.add_coref_links()
		self.add_explicit_links()
		self.block_importance = self.compute_block_importance()

	def add_links(self, nodes):
		levels = dict()
		for n in nodes:
			if not n.is_word():
				if n.depth in levels:
					levels[n.depth].append(n)
				else:
					levels[n.depth] = [n]
		for depth, elements in levels.items():
			self.add_lexical_links(elements, lambda w: w.pos in [POS.ADJ, POS.ADV, POS.NOUN, POS.VERB], OverlapType.CONTENT_OVERLAP)
			self.add_lexical_links(elements, lambda w: w.pos in [POS.NOUN, POS.VERB], OverlapType.TOPIC_OVERLAP)
			self.add_lexical_links(elements, lambda w: w.pos in [POS.NOUN, POS.PRON], OverlapType.ARGUMENT_OVERLAP)
			self.add_semantic_links(elements)

	def add_element(self, element: TextElement) -> List[TextElement]:
		self.graph.add_node(element)
		result = [element]
		if not element.is_sentence():
			for child in element.components:
				result += self.add_element(child)
				self.graph.add_edge(child, element, type=EdgeType.PART_OF)
			self.graph.add_edges_from(zip(element.components[:-1], element.components[1:]), type=EdgeType.ADJACENT)
		return result

	def add_semantic_links(self, elements: List[TextElement]):
		for i, a in enumerate(elements[:-1]):
			for b in elements[i+1:]:
				for model in self.models:
					sim = model.similarity(a, b)
					self.graph.add_edge(a, b, type=EdgeType.SEMANTIC, model=model, value=sim)
					self.graph.add_edge(b, a, type=EdgeType.SEMANTIC, model=model, value=sim)
		
	def add_lexical_links(self, elements: List[TextElement], test: Callable[[Word], bool], link_type: OverlapType):
		words = {element: {word.lemma for word in element.get_words() if test(word)} for element in elements}
		for i, a in enumerate(elements[:-1]):
			for b in elements[i+1:]:
				words_a = words[a]
				words_b = words[b]
				weight = len(words_a & words_b) / (1e-5 + len(words_a | words_b))
				self.graph.add_edge(a, b, type=EdgeType.LEXICAL_OVERLAP, model=link_type, value=weight)
				self.graph.add_edge(b, a, type=EdgeType.LEXICAL_OVERLAP, model=link_type, value=weight)
					
	def add_coref_links(self):
		for node in self.graph.nodes():
			if isinstance(node, Block):
				if node.has_coref:
					for cluster in node.coref_clusters:
						for mention in cluster.mentions:
							if mention != cluster.main and mention.container != cluster.main.container:
								edge = self.get_edge(mention.container, cluster.main.container, edge_type=EdgeType.COREF)
								if edge is None:
									self.graph.add_edge(mention.container, cluster.main.container, type=EdgeType.COREF, details=[(mention.text, cluster.main.text)])
								else:
									edge["details"].append((mention.text, cluster.main.text))


	def add_explicit_links(self):
		explicit_links = []

		for node in self.graph.nodes:
			if isinstance(node, Contribution):
				parent = node.get_parent()
				explicit_links.append((node, parent))

		for node, parent in explicit_links:
			self.graph.add_edge(node, parent, type=EdgeType.EXPLICIT)

	def is_coref_edge(self, a: Block, b: Block) -> bool:
		for _, x, _ in self.edges(a, edge_type=EdgeType.COREF):
			if x is b:
				return True

		return False

	def is_explicit_edge(self, a: Block, b: Block) -> bool:
		for _, x, _ in self.edges(a, edge_type=EdgeType.EXPLICIT):
			if x is b:
				return True

		return False

	def compute_block_importance(self) -> Dict[Block, Dict[Block, float]]:
		block_links = [
        	(a, b, edge["value"])
        	for a, nbrsdict in self.graph.adjacency()
			if isinstance(a, Block)
			for b, edges in nbrsdict.items()
			if isinstance(b, Block)
			for edge in edges.values()
			if edge["type"] == EdgeType.SEMANTIC
		]

		mean = np.mean([value for _, _, value in block_links])
		stdev = np.std([value for _, _, value in block_links])

		block_importance = dict()

		for a, b, value in block_links:
			if not (a in block_importance):
				block_importance[a] = dict()

			if self.is_coref_edge(a, b) or self.is_coref_edge(b, a):
				block_importance[a][b] = value
			elif self.is_explicit_edge(a, b) or self.is_explicit_edge(b, a):
				block_importance[a][b] = value
			elif value > mean + stdev:
				block_importance[a][b] = value
			else:
				block_importance[a][b] = 0

		return block_importance

	def compute_importance(self) -> Dict[TextElement, float]:
		similarities = [value for _, _, value in self.edges(None, edge_type=EdgeType.SEMANTIC)]
		mean = np.mean(similarities)
		stdev = np.std(similarities)
		importance = {}
		for node in self.graph.nodes:
			importance[node] = sum([value for _, _, value in self.edges(node, edge_type=EdgeType.SEMANTIC) if value > mean + stdev])
		return importance

	def edges(self, 
			node: Union[TextElement, Tuple[TextElement, TextElement]], 
			edge_type: EdgeType = None, 
			vector_model: Union[VectorModel, OverlapType] = None) -> List[Tuple[TextElement, TextElement, float]]:
		return [(a, b, data["value"] if "value" in data else 0)
			for a, b, data in self.graph.edges(node, data=True) 
			if (edge_type is None or data["type"] is edge_type) and 
			   (vector_model is None or data["model"] is vector_model)
		]


	def get_edge(self, a: TextElement, b: TextElement, edge_type: EdgeType) -> Dict:
		edge = self.graph[a][b]
		for data in self.graph[a][b].values():
			if (data["type"] is edge_type):
				return data
		return None



from rb.processings.pipeline.estimator import Classifier
from rb.processings.pipeline.dataset import Dataset, Task
from typing import List, Dict
from sklearn.neural_network import MLPClassifier

class MLP(Classifier):
    def __init__(self, dataset: Dataset, tasks: List[Task], params: Dict[str, str]):
        super().__init__(dataset, tasks, params)
        self.model = MLPClassifier(hidden_layer_sizes=params["hidden_layer_sizes"])

    @classmethod
    def parameters(cls):
        return {
            "hidden_layer_sizes": [32, 64, 128]
        }


from typing import List

from rb.core.lang import Lang
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType
from rb.complexity.measure_function import MeasureFunction
from rb.cna.cna_graph import CnaGraph

# create all indices
# dependencies need to be putted in function because otherwise circular dependencies happens
def create(lang: Lang, cna_graph: CnaGraph) -> List["ComplexityIndex"]:
    from rb.complexity.word.wd_len import WdLen
    from rb.complexity.word.wd_diff_lemma import WdDiffLemma
    from rb.complexity.word.no_repetitions import NoRepetitions
    from rb.complexity.word.name_entity_enum import NamedEntityONEnum
    from rb.complexity.word.name_entity import NoNamedEntity
    from rb.complexity.word.wd_max_depth_hypernym_tree import WdMaxDpthHypTree
    from rb.complexity.word.wd_avg_depth_hypernym_tree import WdAvgDpthHypTree
    from rb.complexity.word.no_wd_paths_hypernym_tree import NoWdPathsHypTree
    from rb.complexity.word.polysemy import Polysemy
    from rb.complexity.word.wd_syllab import WdSyllab
    from rb.complexity.word.aoa import Aoa
    from rb.complexity.word.aoa_enum import AoaTypeEnum
    from rb.complexity.word.aoe import Aoe
    from rb.complexity.word.aoe_enum import AoeTypeEnum
    from rb.complexity.word.valence import Valence
    from rb.complexity.word.valence_type import ValenceTypeEnum

    indices = []

    indices.append(WdLen(lang, TextElementType.WORD.value, MeasureFunction.AVG))
    indices.append(WdLen(lang, TextElementType.SENT.value, MeasureFunction.STDEV))

    indices.append(WdSyllab(lang, reduce_depth=TextElementType.SENT.value, reduce_function=MeasureFunction.AVG))
    indices.append(WdSyllab(lang, reduce_depth=TextElementType.SENT.value, reduce_function=MeasureFunction.STDEV))

    indices.append(WdSyllab(lang, reduce_depth=TextElementType.BLOCK.value, reduce_function=MeasureFunction.AVG))
    indices.append(WdSyllab(lang, reduce_depth=TextElementType.BLOCK.value, reduce_function=MeasureFunction.STDEV))

    indices.append(WdDiffLemma(lang, reduce_depth=TextElementType.WORD.value, reduce_function=MeasureFunction.AVG))
    
    indices.append(NoRepetitions(lang, window_size=8, reduce_depth=TextElementType.SENT.value, 
                                    reduce_function=MeasureFunction.AVG))
    indices.append(WdMaxDpthHypTree(lang, reduce_depth=TextElementType.WORD.value, 
                                    reduce_function=MeasureFunction.AVG))
    indices.append(WdAvgDpthHypTree(lang, reduce_depth=TextElementType.WORD.value, 
                                    reduce_function=MeasureFunction.AVG))
    indices.append(NoWdPathsHypTree(lang, reduce_depth=TextElementType.WORD.value, 
                                    reduce_function=MeasureFunction.AVG))
    indices.append(Polysemy(lang, reduce_depth=TextElementType.WORD.value, 
                                    reduce_function=MeasureFunction.AVG))

    if lang is Lang.EN:
        for at in AoaTypeEnum:
            indices.append(Aoa(lang, at, reduce_depth=TextElementType.WORD.value, 
                                    reduce_function=MeasureFunction.AVG))
            indices.append(Aoa(lang, at, reduce_depth=TextElementType.WORD.value, 
                                    reduce_function=MeasureFunction.STDEV))
        for at in AoeTypeEnum:
            indices.append(Aoe(lang, at, reduce_depth=TextElementType.WORD.value, 
                                    reduce_function=MeasureFunction.AVG))
            indices.append(Aoe(lang, at, reduce_depth=TextElementType.WORD.value, 
                                    reduce_function=MeasureFunction.STDEV))
        for vt in ValenceTypeEnum:
            indices.append(Valence(lang, vt, reduce_depth=TextElementType.SENT.value, 
                                    reduce_function=MeasureFunction.AVG))
            indices.append(Valence(lang, vt, reduce_depth=TextElementType.BLOCK.value, 
                                    reduce_function=MeasureFunction.AVG))

    for named_ent_type in NamedEntityONEnum:
        indices.append(NoNamedEntity(lang, named_ent_type=named_ent_type, reduce_depth=TextElementType.WORD.value, 
                                        reduce_function=MeasureFunction.AVG))
    
    return indices



from rb.core.lang import Lang
from rb.core.pos import POS
from enum import Enum
from typing import List, Set, Dict


class POSFeatureExtractor():

    
    def __init__(self):
        pass

    """ factory """
    @classmethod
    def create(clas, lang: Lang) -> "RoPOSFeatureExtractor":
        if lang is Lang.RO:
            from rb.core.pos_features.ro_pos_features.ro_pos_feature_extractor import RoPOSFeatureExtractor
            return RoPOSFeatureExtractor.get_instance()
        else:
            return POSFeatureExtractor()

    def create_ud_dict(self) -> Dict[str, Dict[str, List[Dict[Enum, List[Enum]]]]]:
        pass

    def print_ud_dict(self) -> None:
        pass

    def get_pos_features(self, pos: POS, tag: str) -> Dict[Enum, List[Enum]]:
        pass
    
    def get_all_forms_from_lemma(self, lemma: str) -> Dict[str, Dict[Enum, List[Enum]]]:
        pass


import pandas as pd


def load_file(filename):
    df = pd.read_csv(filename, header=0)

    df = df.drop("name", axis=1)
    df = df.drop("class", axis=1)
    columns = df.columns
    X = []
    y = []
    for index in range(df.shape[0]):
        features = []
        if index % 2 == 0:
            this_row = df.iloc[index]
            next_row = df.iloc[index + 1]
            build_features(columns, features, next_row, this_row)
            y.append(0)
        else:
            this_row = df.iloc[index]
            prev_row = df.iloc[index - 1]
            build_features(columns, features, prev_row, this_row)
            y.append(1)
        X.append(features)
    return X, y


def build_features(columns, features, other_row, this_row):
    for column in columns:
        features.append(float(this_row[column]) / float(other_row[column]) * 100)



from enum import Enum, auto
from rb.core.pos_features.pos_feature import POSFeature
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.lang import Lang
from rb.core.pos import POS
from typing import List
import re


class RoForiegnEnum(Enum):
    YES = auto()


class RoPOSFeatureForeign(POSFeature):

    
    _INSTANCE = None

    def __init__(self):
        POSFeature.__init__(self, lang=Lang.RO, name=RoFeaturesName.FGN,
                            values=None, pos_supported=None)
        self.values = set([ro_fgn for _, ro_fgn in RoForiegnEnum.__members__.items()])
        self.pos_supported = set([POS.ADJ, POS.ADP, POS.DET, POS.NOUN, POS.NUM, POS.PROPN, POS.VERB, POS.X])

    @classmethod
    def get_instance(cls) -> "RoPOSFeatureForeign":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeatureForeign()
        return cls._INSTANCE

    def get_values(self, tag: str) -> List[RoForiegnEnum]:
        pattern = '(Foreign=[a-zA-Z,]*){1}'
        matches = re.findall(pattern, tag)
        if len(matches) == 0: return []
        fgns = matches[0].split('=')[1].split(',')
        values = []
        for fgn in fgns:
            if fgn == 'Yes':
                values.append(RoForiegnEnum.YES)
        return values


from enum import Enum, auto
from rb.core.pos_features.pos_feature import POSFeature
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.lang import Lang
from rb.core.pos import POS
from typing import List
import re


class RoStrengthEnum(Enum):
    STRONG = auto()
    WEAK = auto()

class RoPOSFeatureStrength(POSFeature):


    _INSTANCE = None

    def __init__(self):
        POSFeature.__init__(self, lang=Lang.RO, name=RoFeaturesName.STRN,
                            values=None, pos_supported=None)
        self.values = set([ro_strength for _, ro_strength in RoStrengthEnum.__members__.items()])
        self.pos_supported = set([POS.PRON, POS.X])

    @classmethod
    def get_instance(cls) -> "RoPOSFeatureStrength":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeatureStrength()
        return cls._INSTANCE

    def get_values(self, tag: str) -> List[RoStrengthEnum]:
        pattern = '(Strength=[a-zA-Z0-9,]*){1}'
        matches = re.findall(pattern, tag)
        if len(matches) == 0: return []
        nrs = matches[0].split('=')[1].split(',')
        values = []
        for nr in nrs:
            if nr == 'Strong':
                values.append(RoStrengthEnum.STRONG)
            elif nr == 'Weak':
                values.append(RoStrengthEnum.WEAK)
        return values


from typing import Callable, List, Tuple, Union
from threading import Lock

from nltk.corpus import wordnet as wn
from nltk.corpus.reader.wordnet import Synset
from rb.core.lang import Lang
from rb.core.word import Word
from rb.parser.spacy_parser import SpacyParser

lang_dict = {
    Lang.EN: 'eng',
    Lang.NL: 'nld',
    Lang.FR: 'fra',
    Lang.RO: 'ron',
    Lang.IT: 'ita'
}

SynsetLock = Lock()

def get_synsets(word: str, lang: str, pos: str = None) -> List[Synset]:
    SynsetLock.acquire()
    result = wn.synsets(word, pos=pos, lang=lang)
    SynsetLock.release()
    return result

def get_synset_hypernyms(ss: Synset) -> List[Synset]:
    SynsetLock.acquire()
    result = ss.hypernyms()
    SynsetLock.release()
    return result

def compute_similarity(a: str, b: str, lang: Lang, sim: Callable[[Synset, Synset], float]) -> float:
    if lang not in lang_dict:
        return 0
    lang = lang_dict[lang]
    return min([
        sim(syn_a, syn_b)
        for syn_a in get_synsets(a, lang=lang)
        for syn_b in get_synsets(b, lang=lang)],
        default=0)

def path_similarity(a: str, b: str, lang: Lang) -> float:
    return compute_similarity(a, b, lang, wn.path_similarity)

def leacock_chodorow_similarity(a: str, b: str, lang: Lang) -> float:
    return compute_similarity(a, b, lang, wn.lch_similarity)

def wu_palmer_similarity(a: str, b: str, lang: Lang) -> float:
    return compute_similarity(a, b, lang, wn.wup_similarity)

def get_synonyms(word: Union[str, Word], lang: Lang = None, pos: str = None) -> List[str]:
    if isinstance(word, Word):
        pos = word.pos.to_wordnet()
        lang = word.lang
        word = word.lemma
    if lang not in lang_dict:
        return []
    return list({other
        for ss in get_synsets(word, pos=pos, lang=lang_dict[lang]) 
        for other in ss.lemma_names(lang=lang_dict[lang]) 
        if SpacyParser.get_instance().is_dict_word(other, lang)})

def get_hypernyms(word: Union[str, Word], lang: Lang = None, pos: str = None) -> List[str]:
    if isinstance(word, Word):
        pos = word.pos.to_wordnet()
        lang = word.lang
        word = word.lemma
    if lang not in lang_dict:
        return []
    return list({other
        for ss in get_synsets(word, pos=pos, lang=lang_dict[lang]) 
        for parent in get_synset_hypernyms(ss)
        for other in parent.lemma_names(lang=lang_dict[lang]) 
        if SpacyParser.get_instance().is_dict_word(other, lang)})

def get_min_path_to_root(ss: Synset, lang: Lang = None, pos: str = None) -> int:
    parents: List[Synset] = get_synset_hypernyms(ss)
    shortest_path = float('Inf')
    for parent in parents:
        path_length = 1 + get_min_path_to_root(parent, lang=lang, pos=pos)
        if  path_length < shortest_path:
            shortest_path = path_length
    
    if shortest_path == float('Inf'):
        return 0
    return shortest_path

def get_all_paths_lengths_to_root(word: Union[str, Word], lang: Lang = None, pos: str = None) -> List[int]:
    if isinstance(word, Word):
        pos = word.pos.to_wordnet()
        lang = word.lang
        word = word.lemma

    if lang not in lang_dict:
        return []
        
    paths = []
    for ss in get_synsets(word, pos=pos, lang=lang_dict[lang]):
        path_length = get_min_path_to_root(ss, lang=lang, pos=pos)
        paths.append(path_length)
    
    return paths
if __name__ == "__main__":
    print(path_similarity('hond', 'kat', 'nl'))
    print(get_hypernyms('animal', Lang.RO))



from typing import List, Union

from rb.core.block import Block
from rb.core.lang import Lang
from rb.core.sentence import Sentence
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.core.word import Word
from rb.similarity.vector_model import (CorporaEnum, VectorModel,
                                        VectorModelType)
from rb.utils.downloader import download_tags
from rb.core.document import Document

class MetaDocument(TextElement):
    

    def __init__(self, lang: Lang, text: Union[str, List[str]],
                 depth: int = TextElementType.CONV.value,
                 container: TextElement = None):
        if isinstance(text, str):
            sections = text.split("\n\n")
        else:
            sections = text
            text = "\n\n".join(text)
        TextElement.__init__(self, lang=lang, text=text,
                             depth=depth, container=container)
        self.components = [Document(self.lang, section, container=self) for section in sections]
        


#pylint: disable=import-error
import os
#     except RuntimeError as e:
#         print(e)
import pickle
import sys

import absl
import numpy as np
import rb.processings.diacritics.utils as utils
import tensorflow as tf
from rb.core.lang import Lang
from rb.processings.diacritics.BertCNN import (
    BertCNN, categorical_acc, weighted_categorical_crossentropy)
from rb.processings.diacritics.CharCNN import CharCNN
from rb.processings.encoders.bert import BertWrapper
from rb.utils.downloader import check_version, download_model
from tensorflow.keras.models import load_model

# gpus = tf.config.experimental.list_physical_devices('GPU')
# if gpus:
#     try:
#         for gpu in gpus:
#             tf.config.experimental.set_memory_growth(gpu, True)





class DiacriticsRestoration(object):
	"""
	Wrapper for Diacritics restoration
    """

	def __init__(self):
		# load model
		self._load_model()

	# loads best diacritics model, i.e CharCNN + RoBERT-base FN
	def _load_model(self):
		self.bert_wrapper = BertWrapper(Lang.RO, max_seq_len=128, model_name="base")
		if check_version(Lang.RO, ["models", "diacritice", "RoBERT_base+CNN"]):
			download_model(Lang.RO, ["models", "diacritice", "RoBERT_base+CNN"])
		self.char_to_id_dict = pickle.load(open("resources/ro/models/diacritice/RoBERT_base+CNN/char_dict", "rb"))
		model_path = "resources/ro/models/diacritice/RoBERT_base+CNN/model_11_bert1_tr.h5"
		self.model = load_model(model_path, custom_objects={'BertModelLayer': self.bert_wrapper.bert_layer, 'loss':weighted_categorical_crossentropy(np.ones(5), 5).loss, 
   							'categorical_acc': categorical_acc})

	# replace_all: replaces all diacritics(if existing) with model predictions
	# replace_missing: replaces only characters that accept and don't have diacritics with model predictions; keeps existing diacritics
	def process_string(self, string, mode="replace_all"):
		full_diacritics = set("aăâiîsștț")
		explicit_diacritics = set("ăâîșțĂÂÎȘȚ")

		working_string = string.lower()
		clean_string = ""
		# remove everything not in char_to_id_dict
		for s in working_string:
			if s in self.char_to_id_dict.keys():
				clean_string += s

		working_string = clean_string
		working_string = ''.join([utils.get_char_basic(char) for char in working_string])

		diac_count = 0
		for s in working_string:
			if s in full_diacritics:
				diac_count += 1
		# print(diac_count)
		x_dataset = tf.data.Dataset.from_generator(lambda : utils.generator_bert_cnn_features_string(working_string, self.char_to_id_dict, 11, self.bert_wrapper, 10, 280),
						output_types=({'bert_input_ids': tf.int32, 'bert_segment_ids': tf.int32, 'token_ids': tf.int32, 'sent_ids': tf.int32,
										'mask': tf.float32, 'char_windows': tf.int32}, tf.float32),
						output_shapes=({'bert_input_ids':[10, 128], 'bert_segment_ids':[10, 128], 'token_ids':[280],
										'sent_ids': [280], 'mask': [280], 'char_windows': [280, 11]}, [280, 5]))
		x_dataset = x_dataset.batch(1)

		predictions = self.model.predict(x_dataset, steps=(diac_count//280)+1)

		# print(len(predictions[0]))
		filtered_predictions = []
		for index in range(len(predictions[0])):
			if predictions[1][index] == 1:
				filtered_predictions.append(predictions[0][index])
			
		predictions = np.array(filtered_predictions)
		predicted_classes = list(map(lambda x: np.argmax(x), predictions))
		# print(predictions.shape, len(predicted_classes), predicted_classes)
		# sys.exit()
		prediction_index = 0
		
		complete_string = ""
		for orig_char in string:

			lower_orig_char = orig_char.lower()

			if lower_orig_char in full_diacritics:
				if mode == "replace_all":
					new_char = utils.get_char_from_label(utils.get_char_basic(lower_orig_char), predicted_classes[prediction_index])
				
				elif mode == "replace_missing":					
					if lower_orig_char in explicit_diacritics:
						new_char = lower_orig_char
					else:
						new_char = utils.get_char_from_label(utils.get_char_basic(lower_orig_char), predicted_classes[prediction_index])
				
				prediction_index += 1

				if orig_char.isupper():
					new_char = new_char.upper()

			else:
				new_char = orig_char
			complete_string += new_char

		return complete_string



from rb.similarity.vector_model import VectorModel, VectorModelType
from rb.similarity.vector_model_factory import create_vector_model
from rb.core.document import Document
from rb.core.lang import Lang
from rb.core.word import Word
from rb.core.text_element_type import TextElementType
from collections import Counter
from typing import List, Tuple, Dict, Set
import math
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class KeywordExtractor():

    def __init__(self):
        pass

    def get_vector_model(self, lang: Lang = Lang.RO) -> VectorModel:
        global logger
        if lang is Lang.RO:
            vector_model = create_vector_model(
                Lang.RO, VectorModelType.from_str('word2vec'), "readme")
        elif lang is Lang.EN:
            vector_model = create_vector_model(
                Lang.EN, VectorModelType.from_str("word2vec"), "coca")
        elif lang is Lang.ES:
            vector_model = create_vector_model(
                Lang.ES, VectorModelType.from_str("word2vec"), "jose_antonio") 
        elif lang is Lang.RU:
            vector_model = create_vector_model(
                Lang.RU, VectorModelType.from_str("word2vec"), "rnc_wikipedia") 
        else:
            logger.error(
                f'Language {lang.value} is not supported for keywords task')
            vector_model = None
        return vector_model

    def extract_keywords(self, text: str, lang: Lang = Lang.RO, max_keywords: int = 40) -> List[Tuple[float, Word]]:

        vector_model: VectorModel = self.get_vector_model(lang=lang)
        logger.info('Computing keywords...')
        doc: Document = Document(lang=lang, text=text)
        lemma_words, raw_words = [], []
        scores: List[Tuple[float, Word]] = []

        for word in doc.get_words():
            if word.is_content_word() and not word.is_stop:
                lemma_words.append(word.lemma)
                raw_words.append(word)

        counter = Counter(lemma_words)
        taken_words: Set[str] = set()

        # TODO maybe take the max similarity between all words forms - not so impactful though - Teo
        for ww in raw_words:
            lemma = ww.lemma
            freq = 1 + math.log(counter[lemma])
            sim = vector_model.similarity(ww, doc)
            if lemma not in taken_words:
                taken_words.add(lemma)
                scores.append((sim * freq, ww))

        scores = sorted(scores, reverse=True, key=lambda x: x[0])
        return scores[:max_keywords]

    def keywords_heatmap(self, text: str, lang: Lang = Lang.RO, granularity: TextElementType = TextElementType.SENT,
                         max_keywords: int = 40) -> Dict:

        logger.info('Loading model for keywords extraction...')
        keywords: List[Tuple[float, Word]] = self.extract_keywords(
            text=text, lang=lang)
        vector_model: VectorModel = self.get_vector_model(lang=lang)
        doc: Document = Document(lang=lang, text=text)
        logger.info('Computing keywords heatmap...')

        elements, word_scores = {}, {}

        for kw in keywords:
            word_scores[kw[1].lemma] = {}

        if granularity is TextElementType.SENT:
            for i, sent in enumerate(doc.get_sentences()):
                elements[str(i + 1)] = sent.text
                for kw in keywords:
                    word_scores[kw[1].lemma][str(
                        i + 1)] = str(max(vector_model.similarity(kw[1], sent), 0))
        else:
            for i, block in enumerate(doc.get_blocks()):
                elements[str(i + 1)] = block.text
                for kw in keywords:
                    word_scores[kw[1].lemma][str(
                        i + 1)] = str(max(vector_model.similarity(kw[1], block), 0))

        return {
            "data": {
                "elements": elements,
                "heatmap": {
                    "wordScores":
                        word_scores
                }
            },
            "success": True,
            "errorMsg": ""
        }

    def transform_for_visualization(self, keywords: List[Tuple[int, Word]], lang: Lang) -> Dict:

        vector_model: VectorModel = self.get_vector_model(lang=lang)
        edge_list, node_list = [], []

        for i, kw1 in enumerate(keywords):
            for j, kw2 in enumerate(keywords):
                if i != j and vector_model.similarity(kw1[1], kw2[1]) >= 0.3:
                    edge_list.append({
                        "edgeType": "SemanticDistance",
                        "score": str(max(vector_model.similarity(kw1[1], kw2[1]), 0)),
                        "sourceUri": kw1[1].lemma,
                        "targetUri": kw2[1].lemma
                    })

        for kw in keywords:
            node_list.append({
                "type": "Word",
                "uri": kw[1].lemma,
                "displayName": kw[1].lemma,
                "active": True,
                "degree": str(max(0, float(kw[0])))
            })

        return {
            "data": {
                "edgeList": edge_list,
                "nodeList": node_list
            },
            "success": True,
            "errorMsg": ""
        }



import csv
import os
import uuid
from typing import Dict, List, Tuple

import numpy as np
import time, warnings
from matplotlib import pyplot as plt
from scipy import ndimage
from sklearn.base import BaseEstimator, clone
from sklearn.datasets import make_blobs
from sklearn.ensemble import RandomForestClassifier
from sklearn.utils.metaestimators import if_delegate_has_method
from sklearn.preprocessing import StandardScaler
from sklearn import cluster
from itertools import cycle, islice
from rb.cna.cna_graph import CnaGraph
from rb.complexity.complexity_index import ComplexityIndex, compute_indices
from rb.core.document import Document
from rb.core.lang import Lang
from rb.core.text_element_type import TextElementType
from rb.similarity.vector_model import (CorporaEnum, VectorModel,
                                        VectorModelType)
from rb.similarity.vector_model_factory import (VECTOR_MODELS,
                                                create_vector_model)
from rb.utils.rblogger import Logger

logger = Logger.get_logger()

class Clustering:

    def __init__(self):
        pass

    def plot_clustering(self, X_red, labels, title=None):
        x_min, x_max = np.min(X_red, axis=0), np.max(X_red, axis=0)
        X_red = (X_red - x_min) / (x_max - x_min)
        plt.figure(figsize=(6, 4))
        for i in range(X_red.shape[0]):
            plt.text(X_red[i, 0], X_red[i, 1], str(self.y[i]),
                    color=plt.cm.nipy_spectral(labels[i] / 10.),
                    fontdict={'weight': 'bold', 'size': 9})

        plt.xticks([])
        plt.yticks([])
        if title is not None:
            plt.title(title, size=17)
        plt.axis('off')
        plt.tight_layout(rect=[0, 0.03, 1, 0.95])

    def plot_scatter(self, X,  color, alpha=0.5):
        return plt.scatter(X[:, 0],
                       X[:, 1],
                       c=color,
                       alpha=alpha,
                       edgecolor='k')

    def compute_clustering(self):
        categories = ['stats_general.csv', 'stats_literature.csv', 'stats_science.csv']
        values = {0: [], 1: [], 2: []}
        labels = []
        for i, cat in enumerate(categories):
            stats = csv.reader(open(os.path.join('categories_readme/new_stats/', cat),
                 'rt', encoding='utf-8'))
            for j, row in enumerate(stats):   
                if j == 0:  continue
                vs = []
                for v in row[1:]:
                    vs.append(float(v))
                values[i].append(vs)
                labels.append(i)
        
        gen, lit, science  = np.asarray(values[0]), np.asarray(values[1]), np.asarray(values[2])
        all_samples = np.concatenate((gen, lit, science), axis=0)
        labels = np.asarray(labels)
        y = labels
        X = all_samples

        # clusterer = AgglomerativeClustering(n_clusters=3)
        # cluster_labels = clusterer.fit_predict(X)

        # plt.figure(figsize=(12, 4))

        # plt.subplot(131)
        # self.plot_scatter(X, cluster_labels)
        # plt.title("Ward Linkage")
        # for linkage in ('ward', 'average', 'complete', 'single'):
        #     clustering = AgglomerativeClustering(linkage=linkage, n_clusters=3)
        #     clustering.fit(self.X)
        #     self.plot_clustering(self.X, clustering.labels_, "%s linkage" % linkage)
        # update parameters with dataset-specific values


        # normalize dataset for easier parameter selection
        X = StandardScaler().fit_transform(X)

        # ============
        # Create cluster objects
        # ============
        ward = cluster.AgglomerativeClustering(
            n_clusters=3, linkage='ward')
        complete = cluster.AgglomerativeClustering(
            n_clusters=3, linkage='complete')
        average = cluster.AgglomerativeClustering(
            n_clusters=3, linkage='average')
        single = cluster.AgglomerativeClustering(
            n_clusters=3, linkage='single')
        brc = cluster.Birch(n_clusters=3)

        clustering_algorithms = (
            ('Complete Linkage', complete),
            ('Ward Linkage', ward),
            ('Birch', brc)
        )
        plot_num = 1
        colors = np.array(list(islice(cycle(['#377eb8', '#ff7f00', '#4daf4a']),
                                        int(max(y) + 1))))
        for name, algorithm in clustering_algorithms:
            t0 = time.time()

            # catch warnings related to kneighbors_graph
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    message="the number of connected components of the " +
                    "connectivity matrix is [0-9]{1,2}" +
                    " > 1. Completing it to avoid stopping the tree early.",
                    category=UserWarning)
                algorithm.fit(X)

            t1 = time.time()
            if hasattr(algorithm, 'labels_'):
                y_pred = algorithm.labels_.astype(np.int)
            else:
                y_pred = algorithm.predict(X)

            plt.subplot(1, len(clustering_algorithms) + 1, plot_num)
            plt.title(name, size=18)

            
            plt.scatter(X[:, 0], X[:, 1], s=10, color=colors[y_pred])

            plt.xlim(-2.5, 2.5)
            plt.ylim(-2.5, 2.5)
            plt.xticks(())
            plt.yticks(())
            # plt.text(.99, .01, ('%.2fs' % (t1 - t0)).lstrip('0'),
            #         transform=plt.gca().transAxes, size=15,
            #         horizontalalignment='right')
            plot_num += 1

        plt.subplot(1, len(clustering_algorithms) + 1, plot_num)
        plt.title('True Classes', size=18)

       
        plt.scatter(X[:, 0], X[:, 1], s=10, color=colors[y])

        plt.xlim(-2.5, 2.5)
        plt.ylim(-2.5, 2.5)
        plt.xticks(())
        plt.yticks(())
        # plt.text(.99, .01,
        #         transform=plt.gca().transAxes, size=15,
        #         horizontalalignment='right')
        plt.show()


       



import json
import os
import sys
import zipfile
from typing import List, Union

import wget
from rb.core.lang import Lang
from rb.utils.rblogger import Logger

LINKS = {
    Lang.EN: {
        'models': {
            'coca': {
                'link': "https://nextcloud.readerbench.com/index.php/s/TxamWx6Er9G2wDo/download",
                'version': "https://nextcloud.readerbench.com/index.php/s/npzePeQrpHLE8N6/download"
            },
            'tasa': {
                'link': "https://nextcloud.readerbench.com/index.php/s/CSqbbkxm4En2KKE/download",
                'version': "https://nextcloud.readerbench.com/index.php/s/cXikjDcESTtsJDC/download"
            },
            'enea_tasa': {
                'link': "https://nextcloud.readerbench.com/index.php/s/fd8Ss5NtL6yDoYj/download",
                'version': "https://nextcloud.readerbench.com/index.php/s/5zJxGKxsLSN4YRd/download"
            },
        },
        'spacy': {},
        'aoa': 'https://nextcloud.readerbench.com/index.php/s/estDka8fYiSNWzj/download',
        'wordlists': {
            'link': 'https://nextcloud.readerbench.com/index.php/s/XyeiJCSripBWpx7/download',
            'version': 'https://nextcloud.readerbench.com/index.php/s/xyyGMqeLwTkBRms/download'
        }
    },
    Lang.RO: {
        'models': {
            'diacritice': {
                'RoBERT_base+CNN': {
                    "link": "https://nextcloud.readerbench.com/index.php/s/ZG6qSKwELj87pTF/download",
                    "version": "https://nextcloud.readerbench.com/index.php/s/GRsJP7yFreeicaR/download"
                }
            },
            'sentiment': {
                'small': {
                    "link": "https://nextcloud.readerbench.com/index.php/s/QNYRtoCEQgri499/download",
                    "version": "https://nextcloud.readerbench.com/index.php/s/CF5TmPeEKFNib4H/download"
                },
                'base': {
                    "link": "https://nextcloud.readerbench.com/index.php/s/kC98HTJbPmsXxBj/download",
                    "version": "https://nextcloud.readerbench.com/index.php/s/ix5FRQHEbaP8CFw/download"
                },
                'large': {
                    "link": "https://nextcloud.readerbench.com/index.php/s/pHXqLQ8HXar44WT/download",
                    "version": "https://nextcloud.readerbench.com/index.php/s/aEnbRMiscMjRxcL/download"
                },
            },
            
            'readme': {
                "link": "https://nextcloud.readerbench.com/index.php/s/Sj94ysrmDDxX8YH/download",
                "version": "https://nextcloud.readerbench.com/index.php/s/QXd8847qLz5tNAa/download"
            }
        },
        'spacy': {
            'ro_ud_ft_ner': {
                'link': "https://nextcloud.readerbench.com/index.php/s/5mMb98BXkctjXcP/download",
                'version': "https://nextcloud.readerbench.com/index.php/s/gR6zetbDdfnMMEC/download"
            },
            'ud_tags': {
                'link': 'https://nextcloud.readerbench.com/index.php/s/8WEiAWDSP83sBtx/download',
                'version': 'https://nextcloud.readerbench.com/index.php/s/afE4ZYAMoya9Ekp/download'
            }
        },
        'bert':{
            'small': {
                'link': 'https://nextcloud.readerbench.com/index.php/s/WbtmEn9XebfbBFy/download',
                'version': 'https://nextcloud.readerbench.com/index.php/s/zG2qyqbtongNrKW/download'
            },
            'base': {
                'link': 'https://nextcloud.readerbench.com/index.php/s/Hnn68d6QMr8oKxr/download',
                'version': 'https://nextcloud.readerbench.com/index.php/s/mNG8YbSj7QzpSfK/download'
            },
            'large': {
                'link': 'https://nextcloud.readerbench.com/index.php/s/DeH4mBtZFajKRan/download',
                'version': 'https://nextcloud.readerbench.com/index.php/s/MdMKFDFKJ5feMnN/download'
            },
            'multi_cased_base': {
                'link': 'https://nextcloud.readerbench.com/index.php/s/yGY3AoRYWR2McyD/download',
                'version': 'https://nextcloud.readerbench.com/index.php/s/yWetLtdEz2D4i6f/download'
            },
        },
        'wordnet': "https://nextcloud.readerbench.com/index.php/s/7tDka2CSGYeJqgC/download",
        'wordlists': {
            'link': "https://nextcloud.readerbench.com/index.php/s/oDkA8WfA3J9tzX2/download",
            'version':  'https://nextcloud.readerbench.com/index.php/s/ZWJ34FHy5Zwa65F/download'
        },
        'scoring': {
            'link': 'https://nextcloud.readerbench.com/index.php/s/eMri3i2GHLrQZ24/download',
            'version': 'https://nextcloud.readerbench.com/index.php/s/iTETzcskXZGsBWo/download'
        },
        'classifier': {
            'link': 'https://nextcloud.readerbench.com/index.php/s/DKPXSXfXmtC2445/download',
            'version': 'https://nextcloud.readerbench.com/index.php/s/sMMEqkpiMeX4Pbx/download'
        }
    },
    Lang.RU: {
        'models': {
            'rnc_wikipedia': {
                'link': "https://nextcloud.readerbench.com/index.php/s/PJCmkb2focXiMfd/download",
                'version': "https://nextcloud.readerbench.com/index.php/s/d9Jz7XcxRYpJQgq/download"
            }
        },
        'spacy': {
            'ru_ud_ft': {
                'link': "https://nextcloud.readerbench.com/index.php/s/bWCztgwzdnXowc7/download",
                'version': "https://nextcloud.readerbench.com/index.php/s/TsWjd6KoLjypJQC/download"
            }
        }
    },
    Lang.FR: {
        'models': {
            'le_monde_small': {
                'link': "https://nextcloud.readerbench.com/index.php/s/6MsKa9TtmnAxCc4/download",
                'version': "https://nextcloud.readerbench.com/index.php/s/2jLd8ZiCfmMo88X/download"
            },
        },
    },
    Lang.ES: {
        'models': {
            'jose_antonio': {
                'link': "https://nextcloud.readerbench.com/index.php/s/6J9j4NZgmbALxtE/download",
                'version': "https://nextcloud.readerbench.com/index.php/s/D36i64WMJfdRboD/download"
            },
        },
    },
}

logger = Logger.get_logger()

def download_folder(link: str, destination: str):
    os.makedirs(destination, exist_ok=True)     
    filename = wget.download(link, out=destination, bar=wget.bar_thermometer)
    logger.info('Downloaded {}'.format(filename))
    if zipfile.is_zipfile(filename):
        logger.info('Extracting files from {}'.format(filename))
        with zipfile.ZipFile(filename,"r") as zip_ref:
            zip_ref.extractall(destination)
        os.remove(filename)

def download_file(link: str, destination: str):
    os.makedirs(destination, exist_ok=True)
    filename = wget.download(link, out=destination, bar=wget.bar_thermometer)
    logger.info('Downloaded {}'.format(filename))

def download_model(lang: Lang, name: Union[str, List[str]]) -> bool:
    if isinstance(name, str):
        name = ['models', name]
    if not lang in LINKS:
        logger.info('{} not supported.'.format(lang))
        return False
    path = "/".join(name)
    root = LINKS[lang]
    for key in name:
        if key not in root:
            logger.info('Remote path not found {} ({}).'.format(path, key))
            return False
        root = root[key]
    logger.info("Downloading model {} for {} ...".format(path, lang.value))
    link = root['link'] if isinstance(root, dict) else root
    folder = "resources/{}/{}".format(lang.value, "/".join(name[:-1]))
    download_folder(link, folder)
    return True
    
# TODO check version?
def download_wordlist(lang: Lang) -> bool:
    base_path = os.path.join('resources', lang.value)
    path = os.path.join(base_path, 'wordlists')
    version_path = os.path.join(path, 'version.txt')
    
    if os.path.isfile(version_path) and os.path.isdir(path):
        return True
    if lang not in LINKS:
        logger.info('{} not supported for tags'.format(lang.value))
        return False
    if 'wordlists' not in LINKS[lang]:
        logger.info('No wordlists exist for {}'.format(lang.value))
        return False
    
    link = LINKS[lang]['wordlists']['link']
    logger.info('Downloading wordlists for {} ...'.format(lang.value))
    download_folder(link, base_path)
    logger.info('Downloaded wordlists for {} successfully'.format(lang.value))
    return True

def download_tags(lang: Lang) -> bool:
    path = "resources/{}/{}".format(lang.value, 'spacy')
    full_path = path + '/ud_tags'
    version_path = full_path + '/version.txt'
    if os.path.isfile(version_path):
        return True
    if lang not in LINKS:
        logger.info('{} not supported for tags'.format(lang.value))
        return False
    if 'spacy' not in LINKS[lang]:
        logger.info('Spacy does not exists for {}'.format(lang.value))
        return False
    if 'ud_tags' not in LINKS[lang]['spacy']:
        logger.info('No tags for {}'.format(lang.value))
        return False
    link = LINKS[lang]['spacy']['ud_tags']['link']
    # if you chnage this path you also have to change it in ro_pos_feature_extractor
    logger.info('Downloading tags for {} ...'.format(lang.value))
    
    download_folder(link, path)
    logger.info('Downloaded tags for {} successfully'.format(lang.value))
    return True

def download_spacy_model(lang: Lang, name: str) -> bool:
    return download_model(lang, ['spacy', name])

def download_wordnet(lang: Lang, folder: str) -> bool:
    if lang not in LINKS:
        logger.info('{} not supported.'.format(lang))
        return False
    if 'wordnet' not in LINKS[lang]:
        logger.info('No WordNet found')
        return False
    link = LINKS[lang]['wordnet']
    download_folder(link, folder)

def download_aoa(lang: Lang) -> bool:
    path = "resources/{}/aoa/AoA.csv".format(lang.value)
    if os.path.isfile(path):
        logger.info('File already downloaded')
        return True
    if lang not in LINKS:
        logger.info('{} not supported.'.format(lang))
        return False
    if 'aoa' not in LINKS[lang]:
        logger.info('No AoA found')
        return False
    link = LINKS[lang]['aoa']
    download_file(link, "resources/{}/aoa/".format(lang.value))
    return True

def download_scoring(lang: Lang) -> bool:
    path = "resources/{}/scoring/svr_gamma.p".format(lang.value)
    if os.path.isfile(path):
        logger.info('File already downloaded')
        return True
    if lang not in LINKS:
        logger.info('{} not supported.'.format(lang))
        return False
    if 'scoring' not in LINKS[lang]:
        logger.info('No scoring model found')
        return False
    link = LINKS[lang]['scoring']['link']
    download_file(link, "resources/{}/scoring/".format(lang.value))
    return True


def download_classifier(lang: Lang) -> bool:
    path = "resources/{}/classifier/svr.p".format(lang.value)
    if os.path.isfile(path):
        logger.info('File already downloaded')
        return True
    if lang not in LINKS:
        logger.info('{} not supported.'.format(lang))
        return False
    if 'classifier' not in LINKS[lang]:
        logger.info('No classifier model found')
        return False
    link = LINKS[lang]['classifier']['link']
    download_file(link, "resources/{}/classifier/".format(lang.value))
    return True

def check_spacy_version(lang: Lang, name: str) -> bool:
    return check_version(lang, ['spacy', name])

def check_version(lang: Lang, name: Union[str, List[str]]) -> bool:
    logger.info('Checking version for model {}, {}'.format(name, lang.value))
    if isinstance(name, str):
        name = ['models', name]
    path = "/".join(name)
    folder = "resources/{}/{}".format(lang.value, path)
    try:
        local_version = read_version(folder + "/version.txt")
    except:
        logger.info('Local model {} for {} not found.'.format(path, lang))
        return True
        
    if not lang in LINKS:
        logger.info('{} not supported.'.format(lang))
        return False
    root = LINKS[lang]
    for key in name:
        if key not in root:
            logger.info('Remote path not found {} ({}).'.format(path, key))
            return False
        root = root[key]
    if isinstance(root, dict):
        filename = wget.download(root['version'], out="resources/")
        try:
            remote_version = read_version(filename)
        except:
            logger.info('Error reading remote version for {} ({})'.format(path, lang))
            return False
        return newer_version(remote_version, local_version)
    else:
        logger.info('Could not find version link in links json')
        return True
    
def read_version(filename: str) -> str:
    with open(filename, "r") as f:
        return f.readline()

def newer_version(remote_version: str, local_version: str) -> bool:
    remote_version = remote_version.split(".")
    local_version = local_version.split(".")

    for a, b in zip(remote_version, local_version):
        if int(a) > int(b):
            logger.info('Remote version {} is ahead of local version {}'.format(remote_version, local_version))
            return True
        if int(a) < int(b):
            logger.info('Remote version {} is behind of local version {}'.format(remote_version, local_version))
            return False
    logger.info('Remote version {} is the same as local version {}'.format(remote_version, local_version))
    return False   
    
        

if __name__ == "__main__":
    download_model(Lang.EN, 'coca')



from munkres import Munkres, DISALLOWED, print_matrix

def get_maximum_coupling(author1_details, author2_details, sim_dict):
    author1_articles = author1_details[1]
    author2_articles = author2_details[1]
    # size = max(len(author1_articles), len(author2_articles))

    m = [[DISALLOWED for x in range(len(author2_articles))] for y in range(len(author1_articles))]

    for i1, art1 in enumerate(author1_articles):
        for i2, art2 in enumerate(author2_articles):
            if (art1, art2) in sim_dict:
                if sim_dict[(art1, art2)] > 0.3:
                    m[i1][i2] = (1 - sim_dict[(art1, art2)]) * 100
            elif (art2, art1) in sim_dict:
                if sim_dict[(art2, art1)] > 0.3:
                    m[i1][i2] = (1 - sim_dict[(art2, art1)]) * 100


    mumu = Munkres()
    indexes = mumu.compute(m)
    total = 0
    for row, column in indexes:
        value = 1 - (m[row][column] / 100)
        print(row, column)
        total += value
    print(total)
    return total

if __name__ == "__main__":
    # a1 = ("Mihai Dascalu", ["A1", "A2", "A3"])
    # a2 = ("Profesorul Xavier", ["B1", "B2", "B3", "B4"])
    # s_dict = {
    #     ("A1", "B1"): 0.5, ("A1", "B2"): 0.2, ("A1", "B3"): 0.7, ("A1", "B4"): 0.87,
    #     ("A2", "B1"): 0.25, ("A2", "B2"): 0.49, ("A2", "B3"): 0.81, ("A2", "B4"): 0.33,
    #     ("A3", "B1"): 0.33, ("A3", "B2"): 0.44, ("A3", "B3"): 0.45, ("A3", "B4"): 0.87,
    # }

    a1 = ("Mihai Dascalu", ["A1", "A2"])
    a2 = ("Profesorul Xavier", ["B1", "B2", "B3", "B4"])
    s_dict = {
        ("A1", "B1"): 0.4, ("A1", "B2"): 0.2, ("A1", "B3"): 0.6, ("A1", "B4"): 0.1,
        ("A2", "B1"): 0.5, ("A2", "B2"): 0.1, ("A2", "B3"): 0.3, ("A2", "B4"): 0.7,
    }

    get_maximum_coupling(a1, a2, s_dict)


from gensim.models import Word2Vec 
import spacy
from typing import List, Iterable
from nltk.tokenize import sent_tokenize, WordPunctTokenizer
from os.path import isdir, isfile, join
from os import listdir
from rb.core.lang import Lang
from rb.similarity.vector_model_factory import VectorModelType

from spacy.lang.ro.lemmatizer import LOOKUP
#from spacy.lang.ro.lex_attrs import words

# nlp = spacy.load('en_core_web_sm')
# nlp.remove_pipe('tagger')
# nlp.remove_pipe('parser')
# nlp.remove_pipe('ner')

def split_sentences(fileName: str) -> Iterable[List[str]]:
    # sentences = []
    tokenizer = WordPunctTokenizer()
    with open(fileName, "rt", encoding='utf-8') as f:
        for line in f.readlines():
            for sent in sent_tokenize(line):
                yield [token for token in tokenizer.tokenize(sent)
                       if token.isalpha and not token == '.']

def tokenize_docs(fileName: str) -> Iterable[List[str]]:
    # sentences = []
    tokenizer = WordPunctTokenizer()
    with open(fileName, "rt", encoding='utf-8') as f:
        for line in f.readlines():
            yield [token for token in tokenizer.tokenize(line)
                if token.isalpha and not token == '.']

def load_docs_all(folder: str) -> List[List[str]]:
    all_docs = []
    for f in listdir(folder):
        if isfile(join(folder, f)) and f.endswith(".txt"):
            with open(join(folder, f), "rt", encoding='utf-8', errors='replace') as fin:
                content = fin.read()
                all_docs.append([f, content])
    return all_docs

def load_docs(folder: str) -> Iterable[str]:
    for f in listdir(folder):
        if isfile(join(folder, f)) and f.endswith(".txt"):
            with open(join(folder, f), "r") as fin:
                current = ""
                line = fin.readline()
                while line != "":
                    if line == "\n":
                        yield current
                        current = ""
                    else:
                        current += line
                    line = fin.readline()
                if current != "":
                    yield current

def str_to_lang(s: str) -> Lang:
    s = s.lower()
    if s.strip() == "ro" or s.strip() == "rou" or s.strip() == "romanian":
        return Lang.RO
    elif s.strip() == "en" or s.strip() == "eng" or s.strip() == "english":
        return Lang.EN
    elif s.strip() == "es" or s.strip() == "esp" or s.strip() == "spanish":
        return Lang.ES
    elif s.strip() == "ru" or s.strip() == "rus" or s.strip() == "russian":
        return Lang.RU
    return Lang.EN

def str_to_vmodel(s: str) -> VectorModelType:
    s = s.lower()
    if s.strip() == "lda":
        return VectorModelType.LDA
    elif s.strip() == "lsa":
        return VectorModelType.LSA
    elif s.strip() == "word2vec":
        return VectorModelType.WORD2VEC
    return VectorModelType.WORD2VEC

if __name__ == "__main__":
    pass
    # with open("dict_ro.txt", "wt") as out:
    #     all_words = words | {word for word in LOOKUP} | {lemma for word, lemma in LOOKUP.items()}
    #     all_words = sorted(list(all_words))
    #     for word in all_words:
    #         out.write(word + "\n")




from typing import List, Dict

import spacy

from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.core.word import Word


class Span(TextElement):


    def __init__(self, lang: Lang, text: str, words: List[Word],
         depth: int = TextElementType.SPAN.value):
        super().__init__(lang, text, depth, container=words[0].container)
        self.components = words

    def get_root(self) -> Word:
        return [word for word in self.components 
                if word.head == word or 
                   word.head.index_in_doc < self.components[0].index_in_doc or 
                   word.head.index_in_doc > self.components[-1].index_in_doc
            ][0]
    
    @classmethod
    def from_spacy_span(cls, lang: Lang, spacy_span: spacy.tokens.Span, words: Dict[int, Word]) -> "Word":
        text = spacy_span.text
        our_words = [words[i] for i in range(spacy_span.start, spacy_span.end)]
        return Span(lang=lang, text=text, words=our_words)




from rb.core.lang import Lang
from rb.core.pos import POS
from enum import Enum
from typing import List, Set

""" each pos feature (case, gender, mood, etc.) is defined by this class """

class POSFeature():

    
    def __init__(self, lang: Lang,
                 name: Enum, values: Set[Enum], 
                 pos_supported: Set[POS]):
        self.lang = lang
        self.name = name
        self.values = values
        self.pos_supported = pos_supported

    def get_values(self) -> List:
        pass
    



from typing import List

from rb.core.block import Block
from rb.core.lang import Lang
from rb.core.sentence import Sentence
from rb.core.text_element import TextElement
from rb.core.block import Block
from rb.core.text_element_type import TextElementType
from rb.core.cscl.participant import Participant
from rb.parser.spacy_parser import SpacyParser
from rb.utils.rblogger import Logger

logger = Logger.get_logger()

SIGNIFICANT_LIMIT = 5

class Contribution(TextElement):

    def __init__(self, lang: Lang, text: str,
        participant: Participant,
        parent_contribution: "Contribution",
        timestamp: int,
        depth: int = TextElementType.BLOCK.value,
        container: TextElement = None,
        ):
        
        super().__init__(lang=lang, text=text, depth=depth, container=container)
        self.parent_contribution = parent_contribution
        self.participant = participant
        self.timestamp = timestamp


    def add_sentence(self, sentence: Sentence):
        self.components.append(sentence)
        sentence.container = self

    def get_parent(self) -> "Contribution":
    	return self.parent_contribution

    def get_participant(self) -> Participant:
        return self.participant

    def get_timestamp(self) -> int:
    	return self.timestamp

    def is_significant(self):
        return len(self.get_words()) >= SIGNIFICANT_LIMIT

    def __str__(self):
        return NotImplemented




from typing import Dict, List

import numpy as np
import tensorflow as tf
from rb.core.lang import Lang
from rb.processings.encoders.bert import BertWrapper
from rb.processings.pipeline.dataset import Dataset, Task
from rb.processings.pipeline.estimator import Classifier
from tensorflow import keras


class BertClassifier(Classifier):

    def __init__(self, dataset: Dataset, tasks: List[Task], params: Dict[str, object]):
        super().__init__(dataset, tasks, params)
        self.max_seq_length = 128
        self.bert = BertWrapper(dataset.train_docs[0].lang, max_seq_len=self.max_seq_length)
        self.tasks = tasks
        self.model = self.create_model()
        self.bert.load_weights()
        
    def create_model(self) -> keras.Model:
        inputs, bert_output = self.bert.create_inputs_and_model()
        cls_output = self.bert.get_output(bert_output, "cls")
        features = tf.keras.layers.Input(shape=(len(self.dataset.features),), dtype=tf.float32, name="features")
        outputs = []
        global_hidden = tf.keras.layers.Dense(128)
        for i, task in enumerate(self.tasks):
            masked_features = keras.layers.Lambda(lambda x: x * task.mask)(features)
            concat = keras.layers.concatenate([cls_output, masked_features])
            hidden = global_hidden(cls_output)
            output = keras.layers.Dense(len(task.classes), activation='softmax')(hidden)
            outputs.append(output)
        model = keras.Model(inputs=inputs + [features], outputs=outputs)
        optimizer = keras.optimizers.Adam(learning_rate=1e-6)
        model.compile(optimizer=optimizer, loss='sparse_categorical_crossentropy', metrics=['accuracy'])
        return model



    def train(self):
        train_features = [[doc.indices[feature] for feature in self.dataset.features]
                          for doc in self.dataset.train_docs]
        train_inputs = self.bert.process_input(doc.text for doc in self.dataset.train_docs)
        train_inputs.append(np.array(train_features))
        train_outputs = [np.array(task.get_train_targets()) for task in self.tasks]
        dev_features = [[doc.indices[feature] for feature in self.dataset.features]
                          for doc in self.dataset.dev_docs]
        dev_inputs = self.bert.process_input(doc.text for doc in self.dataset.dev_docs)
        dev_inputs.append(np.array(dev_features))
        dev_outputs = [np.array(task.get_dev_targets()) for task in self.tasks]
        
        self.model.fit(train_inputs, train_outputs, batch_size=32, epochs=10, validation_data=[dev_inputs, dev_outputs])
